import json
import logging
import re
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime
from typing import ClassVar, cast

from dateutil.parser import parse
from rich import box
from rich.console import Console, ConsoleOptions, RenderResult
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from epstein_files.documents.communication import Communication
from epstein_files.documents.document import CLOSE_PROPERTIES_CHAR, EXCERPT_STYLE
from epstein_files.documents.documents.categories import Uninteresting
from epstein_files.documents.config.doc_cfg import DEFAULT_TRUNCATE_TO, NO_TRUNCATE, SHORT_TRUNCATE_TO, DebugDict, EmailCfg, Metadata
from epstein_files.documents.doj_file import DojFile
from epstein_files.documents.emails.constants import *
from epstein_files.documents.emails.email_header import (EMAIL_SIMPLE_HEADER_REGEX,
     EMAIL_SIMPLE_HEADER_LINE_BREAK_REGEX, EmailHeader)
from epstein_files.documents.emails.emailers import extract_emailer_names
from epstein_files.documents.other_file import OtherFile
from epstein_files.people.interesting_people import EMAILERS_OF_INTEREST_SET
from epstein_files.output.epstein_highlighter import highlighter, temp_highlighter
from epstein_files.output.highlight_config import HIGHLIGHTED_NAMES, get_style_for_name
from epstein_files.output.html.builder import table_to_html
from epstein_files.output.html.elements import to_em
from epstein_files.output.layout_elements.file_display import FileDisplay, JustifyMethod
from epstein_files.output.rich import DEFAULT_TABLE_KWARGS, build_table, create_hyperlinks, join_texts, styled_key_value
from epstein_files.util.constant.strings import APPEARS_IN, ARCHIVE_LINK_COLOR, REDACTED, TIMESTAMP_DIM
from epstein_files.util.constant.urls import URL_SIGNIFIERS
from epstein_files.people.names import sort_names
from epstein_files.util.constants import CONFIGS_BY_ID
from epstein_files.util.env import args, site_config
from epstein_files.util.helpers.data_helpers import (AMERICAN_TIME_REGEX, TIMEZONE_INFO, flatten,
     prefix_keys, remove_timezone, uniquify)
from epstein_files.util.helpers.link_helper import link_text_obj
from epstein_files.util.helpers.string_helper import capitalize_first, collapse_newlines, is_bool_prop, strip_pdfalyzer_panels
from epstein_files.util.logging import logger

# Email bod regexes
BAD_FIRST_LINE_REGEX = re.compile(r'^(>>|Grant_Smith066474"eMailContent.htm|LOVE & KISSES)$')
BAD_LINE_REGEX = re.compile(r'^([>=];?|[>»]*=20|\d{1,2}|PAGE INTENTIONALLY LEFT BLANK|Classification: External Communication|Hide caption|Importance:?\s*High|[iI,•]|[1i] (_ )?[il]|, [-,]|L\._|_filtered|si.nature.asc|.*(yiv0232|font-family:|margin-bottom:).*)$')
BAD_SUBJECT_CONTINUATIONS = ['orwarded', 'Hi ', 'Sent ', 'AmLaw', 'Original Message', 'Privileged', 'Sorry', '---']
LINK_LINE_REGEX = re.compile(r"^[>• ]*htt")
LINK_LINE2_REGEX = re.compile(r"^[-\w.%&=/]{5,}$")
QUOTED_REPLY_LINE_REGEX = re.compile(r'(\nFrom:(.*)|wrote:)\n', re.IGNORECASE)
REPLY_TEXT_REGEX = re.compile(rf"^(.*?){REPLY_LINE_PATTERN}", re.DOTALL | re.IGNORECASE | re.MULTILINE)
XML_PLIST_REGEX = re.compile(r"[<=]?\?xml version.*</(plist|xml)>", re.DOTALL)

# Timestamp regexes
BAD_TIMEZONE_REGEX = re.compile(fr'\((UTC|GMT\+\d\d:\d\d)\)|{REDACTED}')
DATE_HEADER_REGEX = re.compile(r'(?:Date|Sent):? +(?!by|from|to|via)([^\n]{6,})\n')
TIMESTAMP_LINE_REGEX = re.compile(r"\d+:\d+")

# numbers
MAX_NUM_HEADER_LINES = 14
MAX_QUOTED_REPLIES = 1
NUM_WORDS_IN_LAST_QUOTE = 6

# TODO: Copy replace_text_with?
DERIVED_CFG_PROPS_TO_COPY = [
    'author_uncertain',
    'category',
    'is_in_chrono',
    'is_interesting',
]

# Junk mail
JUNK_EMAILERS = [
    contact.name
    for junk_hg in HIGHLIGHTED_NAMES if junk_hg.label == Uninteresting.JUNK
    for contact in junk_hg.contacts
]

BCC_LISTS = JUNK_EMAILERS + MAILING_LISTS
TRUNCATE_EMAILS_BY = BCC_LISTS + TRUNCATE_EMAILS_FROM
REWRITTEN_HEADER_MSG = "(janky OCR header fields were prettified, check source if something seems off)"

# TODO: add other forward patterns
REPLY_SPLITTERS = [f"{field}:" for field in COMMON_HEADER_FIELDS] + [
    '********************************',
    'Begin forwarded message',
]

OCR_REPAIRS: dict[str | re.Pattern, str] = {
    # '­': '-',
    # '‐': '-',
    '-­‐': '-',  # TODO: weird hyphens in 027004 and other files that rich doesn't handle well
    re.compile('»'): '>>',
    re.compile(r'grnail\.com'): 'gmail.com',
    'Newsmax. corn': 'Newsmax.com',
    re.compile(r"^(From|To)(: )?[_1.]{5,}", re.MULTILINE): rf"\1: {REDACTED}",  # Redacted email addresses
    # These 3 must come in this order!
    re.compile(r'([/vkT]|Ai|li|(I|7)v)rote:'): 'wrote:',
    re.compile(r"([<.=_HIM][<>.=_HIM14]{5,}[<>.=_HIM]|MOMMINNEMUMMIN) *(wrote:?)?"): rf"{REDACTED} \2",
    re.compile(r"([,<>_]|AM|PM)\n(>)? ?wrote:?"): r'\1\2 wrote:',
    # Headers
    re.compile(r"I ?(od|nl)ine-Images:"): 'Inline-Images:',
    re.compile(r"^((?:B?cc|To):.*)\n(>?;.*)", re.IGNORECASE | re.MULTILINE): r'\1 \2',
    re.compile(r"^From "): 'From: ',
    re.compile(r"^(Sent|Subject) (?![Ff]rom|on|using|[Rr]emote|[Vv]ia|with)", re.MULTILINE): r'\1: ',
    re.compile(r"^Subject[.•]{,2} ", re.MULTILINE): 'Subject: ',
    re.compile(r"^(Forwarded|Original) Message$", re.IGNORECASE | re.MULTILINE): r"--- \1 Message ---",  # Make forward lines match our highlight
    # Excessive quote chars
    re.compile(r"wrote:\n[>»]+(\n[>»]+)"): r"wrote:\1",
    re.compile(r"(^[>»]+\n){2,}", re.MULTILINE): r"\1",
    re.compile(r"\n[<>I ]*wrote:"): ' wrote:',
    # HTML garbage
    re.compile(r'[</=]{,3}(cl|d)?iv>|&[=n]bs[=p];|<=span>|=C\d+\b'): '',
    re.compile(r'^O= ', re.MULTILINE): 'On ',
    re.compile(r"<?=/?[bru]>"): '',
    re.compile(r'(^|\s)[<=][AC]\d+[=>]?', re.MULTILINE): r'\1',
    re.compile(r'[<=]=?/?(br|d?iv)( (class|id)="\w+")?>'): '',
    re.compile(r"\n<mailt.:?(.{,25})[»>]\s*wrote"): r'\1 wrote',
    re.compile(r"^--\w+-- (conversation-id|date-last-viewed).*(flags|remote-id\s\d+)(\s*\d{6,}.*remote-id.*\d+)?", re.MULTILINE): '',
    re.compile(r"<mailto:([-\w=.@]+)[»>]"): r'\1',
    # Names / email addresses
    r'Alireza lttihadieh': ALIREZA_ITTIHADIEH,
    r'bamaby': 'barnaby',
    r'Miroslav Laj6ak': MIROSLAV_LAJCAK,
    r'Ross G°w': ROSS_GOW,
    r'Torn Pritzker': TOM_PRITZKER,
    re.compile(r"( [AP]M,)\s+wrote:$", re.MULTILINE): r'\1 <REDACTED> wrote:',
    re.compile(r' Banno(r]?|\b)'): ' Bannon',
    re.compile(r"\bBamaby\b"): 'Barnaby',
    re.compile(r'gmax ?[1l] ?[@g]ellmax.c ?om'): 'gmax1@ellmax.com',
    re.compile(r"[ijlp']ee[vy]acation[©@a(&,P ]{1,3}g?mail.com"): 'jeevacation@gmail.com',
    'gyahoo.com': '@yahoo.com',
    # Signatures
    re.compile(r'Blac[il]cBerry'): 'BlackBerry',
    'BlackBerry by AT &T': 'BlackBerry by AT&T',
    'BlackBerry from T- Mobile': 'BlackBerry from T-Mobile',
    'Envoy& de': 'Envoyé de',
    'Envoye avec BlackBerry° d': 'Envoye avec BlackBerry® d',
    'from Samsung Mob.le': 'from Samsung Mobile',
    'gJeremyRubin': '@JeremyRubin',
    'Mail for i Phone': 'Mail for iPhone',
    'Sent from Mabfl': 'Sent from Mobile',  # NADIA_MARCINKO signature bad OCR
    'twitter glhsummers': 'twitter @lhsummers',
    re.compile(r"[cC]o-authored with i ?Phone auto-correct"): "Co-authored with iPhone auto-correct",
    re.compile(r"from my ['!()=]([Pp]hone)"): r'from my i\1',
    re.compile(r'from my BlackBerry[0°] wireless device'): 'from my BlackBerry® wireless device',
    re.compile(r'^INW$', re.MULTILINE): REDACTED,
    re.compile(r'Sent from one of my many test mobile devices while on the go and changing the world\s+:\s+so\s+my\s+apologies for any typos'): 'Sent from one of my many test mobile devices while on the go and changing the world : so my apologies for any typos',
    re.compile(r"twitter\.com[i/][lI]krauss[1lt]"): "twitter.com/lkrauss1",
    # links
    'classified-intelligence-\nmichael-flynn-trump': 'classified-intelligence-michael-flynn-trump',
    'on-accusers-rose-\nmcgowan/ ': 'on-accusers-rose-\nmcgowan/\n',
    'the-truth-\nabout-the-bitcoin-foundation/ )': 'the-truth-about-the-bitcoin-foundation/ )\n',
    'woody-allen-jeffrey-epsteins-\nsociety-friends-close-ranks/ ---': 'woody-allen-jeffrey-epsteins-society-friends-close_ranks/\n',
    ' https://www.theguardian.com/world/2017/may/29/close-friend-trump-thomas-barrack-\nalleged-tax-evasion-italy-sardinia?CMP=share btn fb': '\nhttps://www.theguardian.com/world/2017/may/29/close-friend-trump-thomas-barrack-alleged-tax-evasion-italy-sardinia?CMP=share_btn_fb',
    'search-for-secret-putin-\nfortune.html': 'search-for-secret-putin-fortune.html',
    re.compile(r"[h=][t=][t=][p=]://"): 'http://',
    re.compile(r"([h=][t=][t=][op=][s=]|Imps )://"): 'https://',
    re.compile(r'timestopics/people/t/landon jr thomas/inde\n?x\n?\.\n?h\n?tml'): 'timestopics/people/t/landon_jr_thomas/index.html',
    re.compile(r" http ?://www. ?dailymail. ?co ?.uk/news/article-\d+/Troub ?led-woman-history-drug-\n?us ?e-\n?.*html"): '\nhttp://www.dailymail.co.uk/news/article-3914012/Troubled-woman-history-drug-use-claimed-assaulted-Donald-Trump-Jeffrey-Epstein-sex-party-age-13-FABRICATED-story.html',
    re.compile(r"http.*steve-bannon-trump-tower-\n?interview-\n?trumps-\n?strategist-plots-\n?new-political-movement-948747"): "\nhttp://www.hollywoodreporter.com/news/steve-bannon-trump-tower-interview-trumps-strategist-plots-new-political-movement-948747",
    # Subject lines
    r"Arrested in\nInauguration Day Riot": "Arrested in Inauguration Day Riot",
    r"avoids testimony from alleged\nvictims": "avoids testimony from alleged victims",
    r"It's a first, but the buyer's\nanonymous": "It's a first, but the buyer's anonymous",
    r"but\nwatchdogs say probe is tainted": "watchdogs say probe is tainted",
    r"Christmas comes\nearly for most of macro": "Christmas comes early for most of macro",            # 023717
    r"but majority still made good\nmoney because": "but majority still made good money because",      # 023717
    r"COVER UP SEX ABUSE CRIMES\nBY THE WHITE HOUSE": "COVER UP SEX ABUSE CRIMES BY THE WHITE HOUSE",
    r'Priebus, used\nprivate email accounts for': 'Priebus, used private email accounts for',
    r"War on the Investigations\nEncircling Him": "War on the Investigations Encircling Him",
    r"Subject; RE": "Subject: RE",
    r"straining relations between UK and\nAmerica": "straining relations between UK and America",
    re.compile(r"as Putin Mayhem\sTests\sPresident's\sGrip\son\sGOP"): "as Putin Mayhem Tests President's Grip on GOP",
    re.compile(r"deadline re Mr Bradley Edwards vs Mr\s*Jeffrey Epstein", re.I): "deadline re Mr Bradley Edwards vs Mr Jeffrey Epstein",
    re.compile(r"Following Plea That Implicated Trump -\s*https://www.npr.org/676040070", re.I): "Following Plea That Implicated Trump - https://www.npr.org/676040070",
    re.compile(r"for Attorney General -\s+Wikisource, the"): r"for Attorney General - Wikisource, the",
    re.compile(r"JUDGE SWEET\s+ALLOWING\s+STEVEN\s+HOFFENBERG\s+TO\s+TALK\s+WITH\s+THE\s+TOWERS\s+VICTIMS\s+TO\s+EXPLAIN\s+THE\s+VICTIMS\s+SUI\n?T\s+FILING\s+AGAINST\s+JEFF\s+EPSTEIN"): "JUDGE SWEET ALLOWING STEVEN HOFFENBERG TO TALK WITH THE TOWERS VICTIMS TO EXPLAIN THE VICTIMS SUIT FILING AGAINST JEFF EPSTEIN",
    re.compile(r"Lawyer for Susan Rice: Obama administration '?justifiably concerned' about sharing Intel with\s*Trump team -\s*POLITICO", re.I): "Lawyer for Susan Rice: Obama administration 'justifiably concerned' about sharing Intel with Trump team - POLITICO",
    re.compile(r"PATTERSON NEW\s+BOOK\s+TELLING\s+FEDS\s+COVER\s+UP\s+OF\s+BILLIONAIRE\s+JEFF\s+EPSTEIN\s+CHILD\s+RAPES\s+RELEASE\s+DATE\s+OCT\s+10\s+2016\s+STEVEN\s+HOFFENBERG\s+IS\s+ON\s+THE\s+BOOK\s+WRITING\s+TEAM\s*!!!!"): "PATTERSON NEW BOOK TELLING FEDS COVER UP OF BILLIONAIRE JEFF EPSTEIN CHILD RAPES RELEASE DATE OCT 10 2016 STEVEN HOFFENBERG IS ON THE BOOK WRITING TEAM !!!!",
    re.compile(r"PROCEEDINGS FOR THE ST THOMAS ATTACHMENT OF\s*ALL JEFF EPSTEIN ASSETS"): "PROCEEDINGS FOR THE ST THOMAS ATTACHMENT OF ALL JEFF EPSTEIN ASSETS",
    re.compile(r"Subject:\s*Fwd: Trending Now: Friends for three decades"): "Subject: Fwd: Trending Now: Friends for three decades",
    # XML footers
    re.compile('<[iI] ?nteger>'): '<integer>',
    # Misc
    'AVG°': 'AVGO',
    'Saw Matt C with DTF at golf': 'Saw Matt C with DJT at golf',
    re.compile(r"[i. ]*Privileged[- ]*Redacted[i. ]*"): '<PRIVILEGED - REDACTED>',
}

METADATA_FIELDS = [
    'attachments',
    'attachment_file_ids',
    'is_junk_mail',
    'is_mailing_list',
    'recipients',
    'sent_from_device',
]

DEBUG_PROPS = METADATA_FIELDS + [
    'attachment_file_ids',
    'is_persons_first_email',
    'is_word_count_worthy',
]


@dataclass
class Email(Communication):
    """
    Attributes:
        attached_docs (list[OtherFile]): any attachments that exist as `OtherFile` objects
        actual_text (str): Best effort at the text actually sent in this email, excluding quoted replies and forwards.
        derived_cfg (EmailCfg): EmailCfg that was built instead of coming from CONFIGS_BY_ID
        header (EmailHeader): Header data extracted from the text (from/to/sent/subject etc).
        is_persons_first_email (bool): is this the first email we have for this person, whether sender or recipient?
        sent_from_device (str, optional): "Sent from my iPhone" style signature (if it exists).
        signature_substitution_counts (dict[str, int]): Number of times a signature was replaced with
            <...snipped...> per name
        _line_merge_arguments (list[tuple[int] | tuple[int, int]]): preconfigured list of line merges that will fix up
            files from the HOUSE_OVERSIGHT_ collection in memory while leaving the source files untouched
        _was_split_up (bool, optional): True if this file Email was one of the big ones that was split into pieces
            and thus should generally be hidden / not shown
    """
    attached_docs: list[OtherFile] = field(default_factory=list)
    actual_text: str = field(init=False)
    derived_cfg: EmailCfg | None = None
    is_persons_first_email: bool = False
    sent_from_device: str | None = None
    signature_substitution_counts: dict[str, int] = field(default_factory=dict)  # defaultdict breaks asdict :(
    _header: EmailHeader | None = None
    _line_merge_arguments: list[tuple[int] | tuple[int, int]] = field(default_factory=list)
    _was_split_up: bool = False

    # Class variable logging how many headers we prettified while printing, kind of janky
    rewritten_header_ids: ClassVar[set[str]] = set([])

    def __post_init__(self):
        super().__post_init__()
        self.actual_text = self._extract_actual_text()
        self.sent_from_device = self._sent_from_device()

        for signature, name in KNOWN_SIGNATURES.items():
            if self.has_unknown_participant and signature.lower() in self.text.lower():
                self.warn(f"Found known signature for {name} in unattributed email.")

    @property
    def attachment_file_ids(self) -> list[str]:
        """Strings in the Attachments: field in the header, split by semicolon."""
        return [a.file_id for a in self.attached_docs]

    @property
    def attachments(self) -> list[str]:
        """Strings in the Attachments: and Inline-Images: fields in the header, split by semicolon."""
        return self.header.all_attachments

    @property
    def config(self) -> EmailCfg | None:
        """Configured timestamp, if any."""
        if self.file_info.is_local_extract_file:
            this_file_cfg = cast(EmailCfg, CONFIGS_BY_ID.get(self.file_info.file_id))

            # Merge config for file this was extracted from into self.derived_cfg
            if not self.derived_cfg and (extracted_from_cfg := CONFIGS_BY_ID.get(self.file_info.url_slug)):
                self.derived_cfg = this_file_cfg or EmailCfg(id=self.file_id)

                if (extracted_from_description := extracted_from_cfg.complete_description):
                    self.derived_cfg.description = f"{APPEARS_IN} {extracted_from_description}"

                for prop in DERIVED_CFG_PROPS_TO_COPY:
                    derived_cfg_val = getattr(self.derived_cfg, prop)

                    if is_bool_prop(prop) and derived_cfg_val is not None:
                        continue  # Don't overwrite booleans

                    extracted_cfg_val = getattr(extracted_from_cfg, prop)
                    setattr(self.derived_cfg, prop, derived_cfg_val or extracted_cfg_val)

                self.log(f"Constructed synthetic config: {self.derived_cfg}")

            return self.derived_cfg or this_file_cfg
        else:
            return super().config

    @property
    def header(self) -> EmailHeader:
        self._header = self._header or self._extract_header()
        return self._header

    @property
    def html_margin_bottom(self) -> str:
        """Overloaded in `Email` for case of emails with attachments."""
        if self.attached_docs:
            return '6px'
        else:
            return super().html_margin_bottom

    @property
    def info(self) -> list[Text]:
        """Overloads superclass to avoid returning config_description because that's now in the Panel title."""
        if site_config.email_info_in_subtitle:
            return [self.subheader]
        else:
            return super().info

    @property
    def is_fwded_article(self) -> bool:
        """True if this email is just a forward of an article from WSJ or whatever."""
        if self.config:
            if self.config.is_fwded_article is not None:
                return self.config.is_fwded_article
            elif self.config.fwded_text_after:
                return True

        return False

    @property
    def is_interesting(self) -> bool | None:
        """Junk emails are not interesting."""
        if self.is_mailing_list:
            return False
        elif (is_interesting := super().is_interesting) is not None:
            return is_interesting
        elif self.participants.intersection(EMAILERS_OF_INTEREST_SET):
            return True

    @property
    def is_junk_mail(self) -> bool:
        return self.author in JUNK_EMAILERS

    @property
    def is_mailing_list(self) -> bool:
        return self.author in MAILING_LISTS or self.is_junk_mail

    @property
    def is_word_count_worthy(self) -> bool:
        """True if this file should be included in word counts."""
        if self.is_fwded_article:
            return bool(self.config.fwded_text_after) or len(self.actual_text) < 150
        else:
            return not self.is_mailing_list

    @property
    def metadata(self) -> Metadata:
        metadata = super().metadata
        metadata.update(self.truthy_props(METADATA_FIELDS))

        if not self.header.is_empty:
            metadata['header'] = self.header.as_dict()

        if self.attached_docs:
            metadata['attachment_file_ids'] = [f.file_id for f in self.attached_docs]

        return metadata

    @property
    def people(self) -> list[str]:
        """Includes people in attachments."""
        return sorted(uniquify(super().people + flatten([ad.people for ad in self.attached_docs])))

    @property
    def prettified_text(self) -> Text:
        """Cleaned up text ready for printing."""
        should_rewrite_header = self.header.was_initially_empty and self.header.num_header_rows > 0
        num_chars = self._truncate_to_length()
        trim_footer_txt = None
        text = self.text

        # Truncate long emails but leave a note explaining what happened w/link to source document
        if len(text) > num_chars:
            text = text[0:num_chars]
            trim_footer_txt = self.truncation_note(num_chars)

        text = collapse_newlines(text)

        # Rewrite broken headers where the values are on separate lines from the field names
        if should_rewrite_header:
            configured_actual_text = self.config.actual_text if self.config and self.config.actual_text else None
            num_lines_to_skip = self.header.num_header_rows
            lines = []

            # Emails w/configured 'actual_text' are particularly broken; need to shuffle some lines
            if configured_actual_text is not None:
                num_lines_to_skip += 1
                lines += [cast(str, configured_actual_text), '\n']

            lines += text.split('\n')[num_lines_to_skip:]
            text = self.with_header('\n'.join(lines))
            text = _add_line_breaks(text)
            self.rewritten_header_ids.add(self.file_id)

        if args.char_nums:
            text = self._inject_line_numbers(text, args.char_nums)

        lines = [create_hyperlinks(line) for line in text.split('\n')]

        if self.config and self.config.highlighted_pattern:
            logger.debug(f"self.config.highlighted_pattern='{self.config.highlighted_pattern}'")
            txt_highlighter = temp_highlighter(self.config.highlighted_pattern)
        else:
            txt_highlighter = highlighter

        txt = txt_highlighter(join_texts(lines, '\n'))

        if trim_footer_txt:
            txt.append('...\n\n').append(trim_footer_txt)

        return txt

    @property
    def recipients_str(self) -> str:
        return ';'.join([str(r) for r in self.recipients])

    @property
    def subheader(self) -> Text:
        prefix = 'fwded article' if self.is_fwded_article else 'email'
        author_txt = self.author_txt

        if self.config and self.config.is_attribution_uncertain:
            author_txt += Text(f" {QUESTION_MARKS}", style=self.author_style)

        return site_config.email_subheader(prefix, author_txt, self.recipients_txt(), self.timestamp)

    @property
    def subject(self) -> str:
        if self.config and self.config.subject:
            return self.config.subject
        else:
            return self.header.subject or ''

    @property
    def uninteresting_txt(self) -> Text | None:
        """Text to print for uninteresting files."""
        if (uninteresting_txt := super().uninteresting_txt):
            if self.subject:
                uninteresting_txt.append(f' ("{self.subject}")', style='light_yellow3')

            return uninteresting_txt

    @property
    def _summary(self) -> Text:
        """One line summary for logging only."""
        txt = self._summary_with_author

        if len(self.recipients) > 0:
            txt.append(', ').append(styled_key_value('recipients', self.recipients_txt()))

        return txt.append(CLOSE_PROPERTIES_CHAR)

    def file_display(self, align: JustifyMethod | None = None) -> FileDisplay:
        """Allows for proper right vs. left justify."""
        return FileDisplay(
            body_panel=self._body_as_table(self.prettified_text, self.config_description_txt),
            file_info=self.file_id_panel,
            indent=site_config.info_indent,
            justify=align,
            margin_bottom=self.html_margin_bottom,
            subheaders=self.info,
        )

    def is_from_or_to(self, name: str) -> bool:
        """True if `name` is either the author or one of the recipients."""
        return name in self.participants

    def is_note_to_self(self, recipients: list[Name] | None = None) -> bool:
        return (recipients if recipients is not None else self.recipients) == [self.author]

    def recipients_txt(self, max_full_names: int = 2) -> Text:
        """Comma separated colored names (last name only if more than `max_full_names` recipients)."""
        recipients = [r or UNKNOWN for r in self.recipients] if len(self.recipients) > 0 else [UNKNOWN]

        names = [
            Text(r if len(recipients) <= max_full_names else extract_last_name(r), get_style_for_name(r)) + \
                (Text(f" {QUESTION_MARKS}", get_style_for_name(r)) if self.is_recipient_uncertain else Text(''))
            for r in recipients
        ]

        return join_texts(names, join=', ')

    def to_html(self) -> str:
        html = super().to_html()

        if (attachments_table := self._attached_docs_table()):
            indent_props = {'margin-left': to_em(site_config.attachment_indent)}
            html += table_to_html(attachments_table, indent_props)

        return html

    def with_header(self, text: str) -> str:
        """Add the header lines to `text`."""
        return self.header.rewrite_header() + '\n' + text

    def _attached_docs_table(self) -> Table | None:
        if not self.attached_docs:
            return None

        attachments_table_title = f"| Email Attachments for {self.file_info.url_slug}:" # ╏┇┣

        if (doc := self.attached_docs[0]).config and doc.config.show_full_panel:
            if len(self.attached_docs) > 1:
                raise ValueError(f"Can't show more than one panelized attachment for {self}!")

            table = Table(title=attachments_table_title, title_justify='left')
            table.add_column(doc.config.description)
            table.add_row(highlighter(Text(self.attached_docs[0].text, EXCERPT_STYLE)))
            return table
        else:
            return OtherFile.files_preview_table(self.attached_docs, title=attachments_table_title)

    def _body_as_panel(self, text: str | Text, description: Text | None = None) -> Panel:
        """Renders the info info text in the panel's bottom border."""
        return Panel(
            text,
            border_style=self.border_style,
            expand=False,
            title=Text(' ').join(description.split(' ')) if description else None,  # split then join makes rich color subtitle correctly
            title_align='right',
        )

    def _body_as_table(self, text: str | Text, description: Text | None = None) -> Table:
        """Renders the info text as a top row in a table-ish view."""
        panel = Table(
            border_style=self.border_style,
            box=box.ROUNDED,
            header_style='on gray11',
            show_header=bool(description)
        )

        if description and site_config.email_info_in_subtitle:
            description = Text('', justify='right').append(description)

        panel.add_column(description or '')
        panel.add_row(text)
        return panel

    def _debug_props(self) -> DebugDict:
        props = super()._debug_props()
        local_props = self.truthy_props(DEBUG_PROPS)

        if not self.header.is_empty:
            local_props['header'] = self.header.as_dict()
        if self.is_note_to_self():
            local_props['is_note_to_self'] = self.is_note_to_self()

        props.update(prefix_keys(self._debug_prefix, local_props))
        return props

    def _extract_actual_text(self) -> str:
        """The text that comes before likely quoted replies and forwards etc."""
        if self.config and self.config.actual_text is not None:
            return self.config.actual_text

        text = '\n'.join(self.text.split('\n')[self.header.num_header_rows:]).strip()

        if self.config and self.config.fwded_text_after:
            return text.split(self.config.fwded_text_after)[0].strip()
        elif self.header.num_header_rows == 0:
            return self.text

        self.log_top_lines(20, "Raw text:", logging.DEBUG)
        self.log(f"With {self.header.num_header_rows} header lines removed:\n{text[0:500]}\n\n", logging.DEBUG)
        reply_text_match = REPLY_TEXT_REGEX.search(text)

        if reply_text_match:
            actual_num_chars = len(reply_text_match.group(1))
            actual_text_pct = f"{(100 * float(actual_num_chars) / len(text)):.1f}%"
            logger.debug(f"'{self.file_id}': actual_text() reply_text_match is {actual_num_chars:,} chars ({actual_text_pct} of {len(text):,})")
            text = reply_text_match.group(1)

        # If all else fails look for lines like 'From: blah', 'Subject: blah', and split on that.
        for field_name in REPLY_SPLITTERS:
            field_string = f'\n{field_name}'

            if field_string not in text:
                continue

            pre_from_text = text.split(field_string)[0]
            actual_num_chars = len(pre_from_text)
            actual_text_pct = f"{(100 * float(actual_num_chars) / len(text)):.1f}%"
            logger.debug(f"'{self.file_id}': actual_text() fwd_text_match is {actual_num_chars:,} chars ({actual_text_pct} of {len(text):,})")
            text = pre_from_text
            break

        return text.strip()

    def _extract_author(self) -> Name:
        """Overloads superclass method, called at instantiation time."""
        if self.header.author:
            authors = extract_emailer_names(self.header.author)
            return authors[0] if (len(authors) > 0 and authors[0]) else None

    def _extract_header(self) -> EmailHeader:
        """Extract an `EmailHeader` from the OCR text."""
        header_match = EMAIL_SIMPLE_HEADER_REGEX.search(self.text)

        if header_match:
            header = EmailHeader.from_header_lines(header_match.group(0))

            # DOJ file OCR text is broken in a less consistent way than the HOUSE_OVERSIGHT files
            if header.is_empty and not self.file_info.is_doj_file:
                header.repair_empty_header(self.lines)
        else:
            log_level = logging.INFO if self.config else logging.WARNING
            self.log_top_lines(msg='No email header match found!', level=log_level)
            header = EmailHeader(field_names=[])

        logger.debug(f"{self.file_id} extracted header\n\n{header}\n")
        return header

    def _extract_recipients(self) -> list[Name]:
        """Scan the To:, BCC: and CC: fields for known names, falling back to raw strings if no names identified."""
        recipients = flatten([extract_emailer_names(r) for r in self.header.recipients])
        recipients = uniquify(recipients)

        # Assume mailing list emails are to Epstein
        if self.author in BCC_LISTS and (self.is_note_to_self(recipients) or not recipients):
            recipients: list[Name] = [JEFFREY_EPSTEIN]

        # Remove self CCs but preserve self emails
        if not (self.is_note_to_self(recipients) or self.author is None):
            if self.author in self.recipients:
                self.log(f"Removing email to self for {self.author}")

            recipients = [r for r in recipients if r != self.author]

        return sort_names(recipients)

    def _extract_timestamp(self) -> datetime:
        """Find the time this email was sent."""
        if self.header.sent_at and (timestamp := _parse_timestamp(self.header.sent_at)):
            return timestamp

        searchable_lines = self.lines[0:MAX_NUM_HEADER_LINES]
        searchable_text = '\n'.join(searchable_lines)

        if (date_match := DATE_HEADER_REGEX.search(searchable_text)):
            if (timestamp := _parse_timestamp(date_match.group(1))):
                return timestamp

        logger.debug(f"Failed to find timestamp, falling back to parsing {MAX_NUM_HEADER_LINES} lines...")

        for line in searchable_lines:
            if not TIMESTAMP_LINE_REGEX.search(line):
                continue

            if (timestamp := _parse_timestamp(line)):
                logger.debug(f"Fell back to timestamp {timestamp} in line '{line}'...")
                return timestamp

        no_timestamp_msg = f"No timestamp found in '{self.file_path.name}'"

        if self.is_duplicate:
            logger.warning(f"{no_timestamp_msg} but timestamp should be copied from {self.duplicate_of_id}")
        else:
            logger.error(f"{no_timestamp_msg}, top lines:\n" + '\n'.join(self.lines[0:MAX_NUM_HEADER_LINES + 10]))

        return FALLBACK_TIMESTAMP

    def _idx_of_nth_quoted_reply(self, n: int = MAX_QUOTED_REPLIES) -> int | None:
        """Get position of the nth 'On June 12th, 1985 [SOMEONE] wrote:' style line in self.text."""
        header_offset = len(self.header.header_chars)
        text = self.text[header_offset:]

        for i, match in enumerate(QUOTED_REPLY_LINE_REGEX.finditer(text)):
            if i >= n:
                return match.end() + header_offset - 1

    def _merge_lines(self, idx1: int, idx2: int | None = None) -> None:
        """Combine lines numbered 'idx' and 'idx2' into a single line (idx2 defaults to idx + 1)."""
        if idx2 is None:
            self._line_merge_arguments.append((idx1,))
            idx2 = idx1 + 1
        else:
            self._line_merge_arguments.append((idx1, idx2))

        if idx2 < idx1:
            lines = self.lines[0:idx2] + self.lines[idx2 + 1:idx1] + [self.lines[idx1] + ' ' + self.lines[idx2]] + self.lines[idx1 + 1:]
        elif idx2 == idx1:
            raise RuntimeError(f"idx2 ({idx2}) must be greater or less than idx ({idx1})")
        else:
            lines = self.lines[0:idx1]

            if idx2 == (idx1 + 1):
                lines += [self.lines[idx1] + ' ' + self.lines[idx1 + 1]] + self.lines[idx1 + 2:]
            else:
                lines += [self.lines[idx1] + ' ' + self.lines[idx2]] + self.lines[idx1 + 1:idx2] + self.lines[idx2 + 1:]

        self._set_text(lines=lines)

    def _remove_line(self, idx: int) -> None:
        """Remove a line from `self.lines`."""
        num_lines = idx * 2
        self.log_top_lines(num_lines, msg=f'before removal of line {idx}')
        del self.lines[idx]
        self._set_text(lines=self.lines)
        self.log_top_lines(num_lines, msg=f'after removal of line {idx}')

    def _repair(self) -> None:
        """Repair particularly janky files. Note that OCR_REPAIRS are applied *after* other line adjustments."""
        # Some DOJ cleanup needs to happen first if this is a DOJ file.
        super()._repair()

        if self.file_info.is_doj_file:
            new_text = DojFile._remove_bad_lines(strip_pdfalyzer_panels(self.text))
            self._set_text(text=self.repair_ocr_text(DOJ_EMAIL_OCR_REPAIRS, new_text))

        if BAD_FIRST_LINE_REGEX.match(self.lines[0]):
            self._set_text(lines=self.lines[1:])

        # print(self._numbered_lines())
        # import pdb;pdb.set_trace()
        self._set_text(lines=[line for line in self.lines if not BAD_LINE_REGEX.match(line)])
        old_text = self.text

        if self.file_id in LINE_REPAIR_MERGES:
            for merge_args in LINE_REPAIR_MERGES[self.file_id]:
                self._merge_lines(*merge_args)

        if self.file_id in ['025233']:
            self.lines[4] = f"Attachments: {self.lines[4]}"
            self._set_text(lines=self.lines)
        elif self.file_id == '029977':
            self._set_text(text=self.text.replace('Sent 9/28/2012 2:41:02 PM', 'Sent: 9/28/2012 2:41:02 PM'))

        # Bad line removal
        if self.file_id == '025041':
            self._remove_line(4)
            self._remove_line(4)
        elif self.file_id == '029692':
            self._remove_line(3)

        if old_text != self.text:
            self.log(f"Modified text, old:\n\n" + '\n'.join(old_text.split('\n')[0:12]) + '\n')
            self.log_top_lines(12, 'Result of modifications')

        repaired_text = self._repair_links_and_quoted_subjects(self.repair_ocr_text(OCR_REPAIRS, self.text))
        # logger.debug(f"repaired_text\n---\n{self.text}\n---")
        self._set_text(text=repaired_text)
        self._strip_unwanted_text()
        # TODO: we're currently calling this twice to handle lines with HTML garbage that turn into something matching BAD_LINE_REGEX
        self._set_text(lines=[line for line in self.lines if not BAD_LINE_REGEX.match(line)])

    def _repair_links_and_quoted_subjects(self, text: str) -> str:
        """Repair links that the OCR has broken into multiple lines as well as 'Subject:' lines."""
        lines = text.split('\n')
        subject_line = next((line for line in lines if line.startswith('Subject:')), None) or ''
        subject = subject_line.split(':')[1].strip() if subject_line else ''
        new_lines = []
        i = 0

        while i < len(lines):
            line = lines[i]

            if LINK_LINE_REGEX.search(line):
                while i < (len(lines) - 1) \
                        and not (lines[i + 1].startswith('htt') or lines[i + 1].startswith('www') or lines[i + 1].endswith('com')) \
                        and (lines[i + 1].endswith('/') \
                            or any(s in lines[i + 1] for s in URL_SIGNIFIERS) \
                            or LINK_LINE2_REGEX.match(lines[i + 1])):
                    logger.debug(f"{self.filename}: Joining link lines\n   1. {line}\n   2. {lines[i + 1]}\n")
                    line += lines[i + 1]
                    i += 1

                line = line if '[' in line or ']' in line else line.replace(' ', '')
            elif ' http' in line and line.endswith('html'):
                pre_link, post_link = line.split(' http', 1)
                line = f"{pre_link} http{post_link.replace(' ', '')}"
            elif line.startswith('Subject:') and i < (len(lines) - 2) and len(line) >= 40:
                next_line = lines[i + 1]
                next_next = lines[i + 2]

                if len(next_line) <= 1 or any([cont in next_line for cont in BAD_SUBJECT_CONTINUATIONS]):
                    pass
                elif (subject.endswith(next_line) and next_line != subject) \
                        or (HEADER_FIELD_COLON_REGEX.search(next_next) and not HEADER_FIELD_COLON_REGEX.search(next_line)):
                    self.log(f"Fixing broken subject line\n  line: '{line}'\n    next: '{next_line}'\n    next: '{next_next}'\nsubject='{subject}'\n")
                    line += f" {next_line}"
                    i += 1

            new_lines.append(line)
            i += 1

        logger.debug(f"----after line repair---\n" + '\n'.join(new_lines[0:20]) + "\n---")
        return '\n'.join(new_lines)

    def _sent_from_device(self) -> str | None:
        """Find any 'Sent from my iPhone' style signature line if it exist in the 'actual_text'."""
        if (sent_from_match := SENT_FROM_REGEX.search(self.actual_text)):
            sent_from = sent_from_match.group(0).strip()

            if sent_from.startswith('Typos, misspellings courtesy of iPhone'):  # special handling for linda stone periods
                return sent_from.removesuffix('.')
            elif sent_from.startswith('sent'):
                return capitalize_first(sent_from)
            else:
                return sent_from

    def _strip_unwanted_text(self) -> None:
        """Add newlines before quoted replies, snip signatures and XML, etc.."""
        # Insert line breaks now unless header is broken, in which case we'll do it later after fixing header
        # self.(f"text before _add_line_breaks:\n\n{self.text}\n---")
        text = self.text if self.header.was_initially_empty else _add_line_breaks(self.text)
        text = REPLY_REGEX.sub(r'\n\1', text)  # Newlines between quoted replies
        text = FORWARDED_TOO_MUCH_SPACE_REGEX.sub(r'\1\n', text)
        # self.warn(f"text after _add_line_breaks:\n\n{text}\n---")

        for name, signature_regex in EMAIL_SIGNATURE_REGEXES.items():
            signature_replacement = f'<...snipped {name.lower()} email signature...>'
            text, num_replaced = signature_regex.subn(signature_replacement, text)
            self.signature_substitution_counts[name] = self.signature_substitution_counts.get(name, 0)
            self.signature_substitution_counts[name] += num_replaced

        # Remove XML cruft in some files
        text, num_plists_stripped = XML_PLIST_REGEX.subn(XML_STRIPPED_MSG, text)

        if num_plists_stripped:
            self.log(f"Replaced {num_plists_stripped} XML plists...")

        self._set_text(text=collapse_newlines(text).strip())

    def _truncate_to_length(self) -> int:
        """Decide how many chars we should limit the dislpay of this email to."""
        quote_cutoff = self._idx_of_nth_quoted_reply()  # Trim if there's many quoted replies
        includes_truncate_term = next((term for term in TRUNCATE_TERMS if term in self.text), None)
        num_chars: int

        if args.whole_file:
            num_chars = len(self.text)
        elif args.truncate:
            num_chars = args.truncate
        elif self.config and (truncate_at := self.config.truncate_at):
            if self.config.is_excerpt:
                raise ValueError(f"Emails don't support truncate_to as a tuple")

            num_chars = len(self.text) if truncate_at == NO_TRUNCATE else truncate_at
        elif self.author in TRUNCATE_EMAILS_BY \
                or any([self.is_from_or_to(n) for n in TRUNCATE_EMAILS_FROM_OR_TO]) \
                or includes_truncate_term:
            num_chars = min(quote_cutoff or DEFAULT_TRUNCATE_TO, SHORT_TRUNCATE_TO)
        else:
            if quote_cutoff and quote_cutoff < DEFAULT_TRUNCATE_TO:
                trimmed_words = self.text[quote_cutoff:].split()

                if '<...snipped' in trimmed_words[:NUM_WORDS_IN_LAST_QUOTE]:
                    num_trailing_words = 0
                elif trimmed_words and trimmed_words[0] in ['From:', 'Sent:']:
                    num_trailing_words = NUM_WORDS_IN_LAST_QUOTE
                else:
                    num_trailing_words = NUM_WORDS_IN_LAST_QUOTE

                if trimmed_words:
                    last_quoted_text = ' '.join(trimmed_words[:num_trailing_words])
                    num_chars = quote_cutoff + len(last_quoted_text) + 1 # Give a hint of the next line
                else:
                    num_chars = quote_cutoff
            else:
                # TODO: Added some padding to self.length because newlines may be added in prettification but this sucks
                num_chars = min(self.length + 100, DEFAULT_TRUNCATE_TO)

            # Always print whole email for 1st email for actual people
            if self.is_persons_first_email and num_chars < self.length and \
                    not (self.is_duplicate or self.is_fwded_article or self.is_mailing_list):
                self.log(f"{self} Overriding cutoff {num_chars} for first email")
                num_chars = self.length + 100

        log_args = {
            'num_chars': num_chars,
            'is_persons_first_email': self.is_persons_first_email,
            'author_truncate': self.author in TRUNCATE_EMAILS_BY,
            'is_fwded_article': self.is_fwded_article,
            'is_quote_cutoff': quote_cutoff == num_chars,
            'includes_truncate_term': json.dumps(includes_truncate_term) if includes_truncate_term else None,
            'quote_cutoff': quote_cutoff,
        }

        log_args_str = ', '.join([f"{k}={v}" for k, v in log_args.items() if v])
        self.log(f"Truncate determination: {log_args_str}")
        return num_chars

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        prettified_txt = self.prettified_text
        max_line_len = max(*[len(line) for line in prettified_txt.split('\n')])
        panel_subtitle = None

        if self.config_description_txt and site_config.email_info_in_subtitle:
            max_line_len = max(max_line_len, len(self.config_description_txt.plain))
            panel_subtitle = Text('').append(self.config_description_txt)
            panel_subtitle.justify = 'right'

        if args.panelize_emails:
            body = self._body_as_panel(prettified_txt, panel_subtitle)
        else:
            body = self._body_as_table(prettified_txt, panel_subtitle)

        yield self.rich_header()
        body_bottom_padding = 0 if self.attached_docs else 1
        yield Padding(body, (0, 0, body_bottom_padding, site_config.other_files_table_indent))

        if (attachments_table := self._attached_docs_table()):
            yield Padding(attachments_table, (0, 0, 1, site_config.attachment_indent))

    @staticmethod
    def build_emails_table(emails: list['Email'], name: Name = '', title: str = '', show_length: bool = False) -> Table:
        """Turn a list of `Email` objects into a `Table` with sender, recipient, and subject line."""
        if title and name:
            raise ValueError(f"Can't provide both 'author' and 'title' args")
        elif name == '' and title == '':
            raise ValueError(f"Must provide either 'author' or 'title' arg")

        author_style = get_style_for_name(name, allow_bold=False)
        link_style = author_style if name else ARCHIVE_LINK_COLOR
        min_width = len(name or UNKNOWN)
        max_width = max(20, min_width)

        columns = [
            {'name': 'Sent At', 'justify': 'left', 'style': TIMESTAMP_DIM},
            {'name': 'From', 'justify': 'left', 'min_width': min_width, 'max_width': max_width},
            {'name': 'To', 'justify': 'left', 'min_width': min_width, 'max_width': max_width + 2},
            {'name': 'Length', 'justify': 'right', 'style': 'wheat4'},
            {'name': 'Subject', 'justify': 'left', 'min_width': 35, 'style': 'honeydew2'},
        ]

        table = build_table(
            title or None,
            cols=[col for col in columns if show_length or col['name'] not in ['Length']],
            border_style=DEFAULT_TABLE_KWARGS['border_style'] if title else author_style,
            header_style="bold",
            highlight=True,
        )

        for email in emails:
            fields = [
                link_text_obj(email.file_info.external_url, email.timestamp_without_seconds, style=link_style),
                email.author_txt,
                email.recipients_txt(max_full_names=1),
                f"{email.length}",
                email.subject,
            ]

            if not show_length:
                del fields[3]

            table.add_row(*fields)

        return table


def _add_line_breaks(email_text: str) -> str:
    text = EMAIL_SIMPLE_HEADER_LINE_BREAK_REGEX.sub(r'\n\1\n', email_text).strip()
    logger.debug(f"text after EMAIL_SIMPLE_HEADER_LINE_BREAK_REGEX _add_line_breaks()\n---\n{text}\n---")
    return FORWARDED_TOO_MUCH_SPACE_REGEX.sub(r'\1\n', text)


def _parse_timestamp(timestamp_str: str) -> None | datetime:
    try:
        if (american_date_match := AMERICAN_TIME_REGEX.search(timestamp_str)):
            timestamp_str = american_date_match.group(1)
        else:
            timestamp_str = timestamp_str.replace('(GMT-05:00)', 'EST')
            timestamp_str = BAD_TIMEZONE_REGEX.sub(' ', timestamp_str).strip()

        timestamp = parse(timestamp_str, fuzzy=True, tzinfos=TIMEZONE_INFO)
        logger.debug(f'Parsed timestamp "%s" from string "%s"', timestamp, timestamp_str)
        return remove_timezone(timestamp)
    except Exception as e:
        logger.debug(f'Failed to parse "{timestamp_str}" to timestamp!')



EFTA00009622_TEXT = """<REDACTED> s GJ Test.          7/19/2006

<REDACTED> dob          16 yrs

<REDACTED> - Loxahatchee

2005 <REDACTED>          9th grade

<REDACTED> was outside          asked to date <REDACTED>

1/10/19 <REDACTED>

~~talked in depth~~

met for first time the night before went to JE
11:30 ish 10
do you want to make $200

met my friend Jeff  bighouse

but you have to give massage 45 min + that's all you have to do get $200

<REDACTED> told <REDACTED> + <REDACTED> went to Bathroom
to argue
came back out <REDACTED> called Jeff on the phone.
1 assistant answered the pho...

Page 2
Jeff asst answered phone - <REDACTED> had someone to give massage Lady asked how old I was
she told the lady 18 + went to <REDACTED>

Its $200     didn't care what
<REDACTED> - Lady trying to verify 18 Yes
Graduated from <REDACTED>     Known <REDACTED> for a long time
& the next morning

<REDACTED> + another girl     - going to Mall he <REDACTED> didn't care

<REDACTED> asked dad for $     yeh sure fine ok

House in Palm Beach
all got out of car     gate at back door     security guard the saw us     the <REDACTED> asked us where...
Page 3

<REDACTED> said Jeffrey
Went to kitchen     waited 15 min

JE + asst     walked in back door

shook hands     met both JE asst.
(A) JE     chose gonna go first

JE said how about you
<REDACTED>     upstairs     "asst lady"

walked me up there stripping out all this picture

She pulled out massage table me     front of couch     blanket over table     opened drawer     whole bunch of lotions     + pick some out
She told me keep bra + panties on + get undress

No <REDACTED>     told her I thought a massage keep your clothing

Page 4

stay in the room
I came in the room     HE + shook hands
Came back in the room

JE:     old     50ish

initially     guy golfy     khaki pants pullover

t-shirt + jeans     thong     bra

hesitant     thinking what's going on. I didn't care cause I wanted $200

He laid on massage table     He said grab any 3 lotions

Started to give him

What M.S.     did you go to told     him she was 16 Not     do you like to do     small talk <REDACTED>

Sat on table     15-20 minute conversation     another house in NY

Page 5

He asked if I wanted to make an extra $300
Straddle him

massage over -
but if you're not comfortable giving a massage can I     <REDACTED>     on you     ~~anything~~
<REDACTED>     said ok
JE left + came ba

<REDACTED>     kind of like a <REDACTED>
grabbed that + layed down back on table     ~~old~~ ~~old~~ hands

Give me a     massage on chest use the <REDACTED>     10 minutes or so ( after     a couple ~~of~~ minutes ( on <REDACTED>
(A) <REDACTED> said if he did that he would have to pay her extra money if he >
Page 6

<REDACTED>     not she would have to didn't     tell her     ~~get~~ ~~over~~ do her clothes at He <REDACTED>     you get paid

<REDACTED>

Yes     Saw <REDACTED>

He opened wallet     got the   $200

He told her to write Name + #

after that left the room

I didn't see him <REDACTED>

I was scared to tell

<REDACTED>     her body not his body
Went down by herself    got 1st  No didn't do anything  but <REDACTED>  $200 ~~too~~

Page 7
I figure it was probably because she brought me

What happened     I got $300 I was excited

What did you have to do with him
If you do this you get that
Laughing ~~in~~     Thanks for fully me I was going to ~~much~~ ~~go~~ ~~many~~ Oh sorry.
Didn't plan it

If you ever need money
Oh yeh     - everybody

I did not ever go back to house

at school my best friend I told her if you ever want to make 300.-o massage Told <REDACTED>     <REDACTED>

Page 8
<REDACTED>     telling everybody I confronted her     altercation get into a fight <REDACTED>

Spreading rumors about me Let me see in your purse you don't have to lie Stefand I worked at <REDACTED>

I didn't think it was her business and     why would tell on myself

You don't have to lie I knew what happen
Everything came out

<REDACTED>     then I didn't want to tell them what had happened,

I <REDACTED>

<REDACTED> Family Setty a lot problems

No     the only <REDACTED>     sometimes
Page 9

<REDACTED>     Have you ever da[ting] <REDACTED> Survey -     Yeh
<REDACTED>     not anymore
14 birthday     my dad
I did <REDACTED>     lots of time

$250,000     its a joke
tall skinny     dark hair
Page 10
""".strip()



from epstein_files.documents.documents.categories import CONSTANT_CATEGORIES, Interesting, Neutral
from epstein_files.documents.config.config_builder import FBI_REPORT, fbi_defense_witness, fbi_report, letter
from epstein_files.documents.config.doc_cfg import NO_TRUNCATE, DocCfg, EmailCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *
from epstein_files.util.helpers.string_helper import join_truthy, quote
from epstein_files.util.logging import logger


GOVERNMENT_CFGS = [
    DocCfg(
        id='024117',
        description=f"anti-money laundering (AML), Bank Secrecy Act (BSA), & terrorist financing (CFT) US law FAQ",
        is_interesting=True,
    ),
    DocCfg(
        id='EFTA00315076',
        date='2008-06-01',
        date_uncertain=True,
        description=f"visitor list during Epstein's 2009 incarceration in {PALM_BEACH}",
        show_full_panel=True,
    ),
    DocCfg(
        id='EFTA00306074',
        date='2008-06-01',
        date_uncertain=True,
        description=f"approved mail list during Epstein's 2009 incarceration in {PALM_BEACH}",
        show_full_panel=True,
    ),
    DocCfg(id='EFTA00109783', author=BUREAU_OF_PRISONS, description='prisoner assignments', date='2019-08-03'),
    DocCfg(id='EFTA00035921', author=BUREAU_OF_PRISONS, description="Lieutenant's Logs", date='2019-08-06'),
    DocCfg(id='EFTA00120617', author=BUREAU_OF_PRISONS, description='"Prisoner Remand or Order to Deliver" forms', date='2019-07-30'),
    DocCfg(id='EFTA00109163', author=BUREAU_OF_PRISONS, description='various Metropolitan Correctional Center forms showing Konstantin Ignatov', date='2019-08-08', is_interesting=True),
    DocCfg(id='EFTA00109654', author=BUREAU_OF_PRISONS, description='roster of inmates at Metropolitan Correctional Center', date='2019-08-08'),
    DocCfg(id='EFTA00108533', author=BUREAU_OF_PRISONS, description='roster of inmates at Metropolitan Correctional Center', date='2019-07-23'),
    DocCfg(id='EFTA00108552', author=BUREAU_OF_PRISONS, description='roster of inmates at Metropolitan Correctional Center', date='2019-07-23'),
    DocCfg(id='EFTA00039153', author=BUREAU_OF_PRISONS, description='List of Exhibits, Chapter 2', date='2019-01-06'),
    DocCfg(id='EFTA00039025', author=BUREAU_OF_PRISONS, description="report on death of Jeffrey Epstein", is_interesting=True),
    DocCfg(id='EFTA00039190', author=BUREAU_OF_PRISONS, description='Special Housing Units', date='2016-11-23', is_interesting=False),
    DocCfg(id='EFTA00120459', author=BUREAU_OF_PRISONS, replace_text_with='handwritten log of prisoner movements', date='2019-08-09'),
    DocCfg(id='EFTA00039227', author=BUREAU_OF_PRISONS, replace_text_with='Inmate Discipline Program Statement'),
    DocCfg(id='EFTA00039295', author=BUREAU_OF_PRISONS, replace_text_with='Inmate Telephone Privileges Program Statement'),
    DocCfg(id='EFTA00039312', author=BUREAU_OF_PRISONS, replace_text_with='Program Statement / Memo about BOP Pharmacy Program'),
    DocCfg(id='EFTA00039351', author=BUREAU_OF_PRISONS, replace_text_with='Program Statement / Memo about BOP Pharmacy Program'),
    DocCfg(id='EFTA00039156', author=BUREAU_OF_PRISONS, replace_text_with='Standards of Employee Conduct'),
    DocCfg(
        id='EFTA00164939',
        author=DOJ,
        date='2025-09-01',
        date_uncertain='approximate',
        description='Powerpoint summary of Child Sex Trafficking Task Force Epstein investigation',
        is_interesting=True,
    ),
    DocCfg(id='EFTA02731200', author=DOJ, description="memo about potential prosecution of Epstein's assistant"),
    DocCfg(id='EFTA02731082', author=DOJ, description="memo about investigation into Epstein's co-conspirators"),
    DocCfg(id='EFTA02730741', author=DOJ, date='2025-05-01', date_uncertain=True, description="Evidence list for 50D-NY-3027571 Filtering On 'Type(s): 1B'"),
    DocCfg(id='EFTA02730486', author=DOJ, date='2025-05-01', date_uncertain=True, description="Evidence list for 50D-NY-3027571 Filtering On '1A'"),
    DocCfg(id='EFTA00040006', author=DOJ, date='2019-08-27', description='Personal History of Defendant Jeffrey Epstein + grand jury indictment'),
    DocCfg(id='EFTA02731226', author=DOJ, date='2021-03-14', description=f'memo seeking authorization to charge {GHISLAINE_MAXWELL} with additional offenses'),
    DocCfg(
        id='EFTA01246710',
        author=FBI,
        description="interview where Epstein's chef says Donald Trump came to Epstein's house for dinner",
        truncate_to=(6000, 7500),
    ),
    DocCfg(
        id='EFTA00174375',
        author=FBI,
        date='2022-11-18',
        description=f"interview of {LUKE_D_THORBURN} with a lot of takes on Epstein, China, and {STEVE_BANNON}",
    ),
    DocCfg(
        id='EFTA00159321',
        author=FBI,
        description='interview re: Paolo Zampolli, Epstein assaults, and the possibility Epstein introduced Melania to Donald Trump',
    ),
    # https://www.bbc.com/news/articles/c6271ngl014o
    DocCfg(
        id='EFTA01660676',
        author=FBI,
        description='tip about recently convicted rapists Tal and Oren Alexander at Epstein\'s house',
        show_full_panel=True,
    ),
    DocCfg(id='EFTA02858481', author=FBI, description='interview with victim Jennifer Aros'),
    DocCfg(id='EFTA00101927', author=FBI, description=f'interview with someone who mentions au pair of Glenn and {EVA_DUBIN} held against her will'),
    DocCfg(id='EFTA00247131', author=FBI, description='search warrant for New York house', date='2019-07-07'),
    DocCfg(id='EFTA01249591', author=FBI, description=f"allegations against {HENRY_JARECKI}"),
    DocCfg(id='EFTA00081226', author=FBI, description='interview with minor victim'),
    DocCfg(id='EFTA00038915', author=FBI, description='interview with minor victim who said Epstein knew she was 14'),
    DocCfg(id='EFTA00023055', author=FBI, description="evidence of notes left about newly recruited underage girls by girls giving massages"),
    DocCfg(id='EFTA00222943', author=FBI, description=f"agent believes computers were removed from Epstein's residence"),
    DocCfg(id='EFTA00020506', author=FBI, description='"chauffeur told Epstein \'I have something on you, remember what I buried!\'"', show_full_panel=True),
    DocCfg(id='EFTA00084614', author=PALM_BEACH_POLICE, description='incident report detailing the investigation into Jeffrey Epstein'),
    DocCfg(id='EFTA00007893', author=PALM_BEACH_POLICE, description=f"receipts, notes, bank statements of {GHISLAINE_MAXWELL}"),
    DocCfg(id='EFTA00005569', author=PALM_BEACH_POLICE, replace_text_with='photo lineup featuring Epstein', date='2005-03-17'),
    DocCfg(id='EFTA01227877', description='multi entry visa for the Russian Federation', date='2018-06-25', show_full_panel=True),
    DocCfg(id='EFTA00128379', description='analysis of two of Epstein\'s desktop computers', date='2019-01-06', is_interesting=True),
    DocCfg(id='EFTA02730274', description='evidence inventory that appears to have since been deleted from the DOJ website'),
    DocCfg(id='EFTA00263284', description='notes about deceit and sexual manipulation by Australian professor Vincent Bulone', is_interesting=True),
    DocCfg(id='EFTA00001884', description='photo of letter from Virgin Islands DOJ to Epstein', date='2019-03-14'),
    DocCfg(id='EFTA00074744', description="USVI court filing about Epstein will and estate"),
    DocCfg(id='EFTA00007157', description='victim list and police log'),
    fbi_defense_witness('EFTA02730267', 'Malcolm Grumbridge', '2022-04-14'),
    fbi_defense_witness('EFTA02730271', REDACTED, '2022-03-22'),
    fbi_defense_witness('EFTA02730477', 'Roderic Alexander', '2022-01-19'),
    letter('EFTA00098456', PAUL_G_CASSELL, ['Scotland Yard'], 'International Sex Trafficking by Jeffrey Epstein, contains court filings'),

    # Emails
    EmailCfg(
        id='EFTA00129096',
        date='2025-04-03 7:13:35 PM',
        description=f'background check on Tim Draper and {MASHA_DROKOVA}',
        show_full_panel=True,
        truncate_to=NO_TRUNCATE,  # TODO this shouldn't be necessary?
    ),
    EmailCfg(id='EFTA02730483', author=FBI, date='2023-07-11T08:25:00', date_uncertain='actually reply timestamp'),
    EmailCfg(id='EFTA02731552', author=FBI, recipients=[USANYS], date='2021-05-26 16:12:00', recipient_uncertain=True),
    EmailCfg(id='EFTA00039971', author=FBI, recipients=[USANYS], recipient_uncertain=True),
    EmailCfg(id='EFTA00037683', description=f"tip that the murder of DC Madam Jeanne Palfrey might be connected to Epstein's network"),
]



from pathlib import Path
from subprocess import run

from rich.text import Text

from epstein_files.documents.document import Document
from epstein_files.output.rich import console
from epstein_files.util.constant.strings import HOUSE_OVERSIGHT_PREFIX
from epstein_files.util.env import DOCS_DIR
from epstein_files.util.helpers.file_helper import extract_file_id


def diff_files(files: list[str]) -> None:
    """Diff the contents of two `Document` objects after all cleanup, BOM removal, etc."""
    if len(files) != 2:
        raise RuntimeError('Need 2 files')
    elif files[0] == files[1]:
        raise RuntimeError(f"Filenames are the same!")

    files = [f"{HOUSE_OVERSIGHT_PREFIX}{f}" if len(f) == 6 else f for f in files]
    files = [f if f.endswith('.txt') else f"{f}.txt" for f in files]
    tmpfiles = [Path(f"tmp_{f}") for f in files]
    docs = [Document(DOCS_DIR.joinpath(f)) for f in files]

    for i, doc in enumerate(docs):
        doc._write_clean_text(tmpfiles[i])

    cmd = f"diff {tmpfiles[0]} {tmpfiles[1]}"
    console.print(f"Running '{cmd}'...")
    results = run(cmd, shell=True, capture_output=True, text=True).stdout

    for line in _color_diff_output(results):
        console.print(line, highlight=True)

    console.print(f"Possible suppression with: ")
    console.print(Text('   suppress left: ').append(f"   '{extract_file_id(files[0])}': 'the same as {extract_file_id(files[1])}',", style='cyan'))
    console.print(Text('  suppress right: ').append(f"   '{extract_file_id(files[1])}': 'the same as {extract_file_id(files[0])}',", style='cyan'))

    for f in tmpfiles:
        f.unlink()


def _color_diff_output(diff_result: str) -> list[Text]:
    txts = [Text('diff output:')]
    style = 'dim'

    for line in diff_result.split('\n'):
        if line.startswith('>'):
            style='spring_green4'
        elif line.startswith('<'):
            style='sea_green1'

        txts.append(Text(line, style=style))

    return txts



import gzip
import json
import pickle
import re
import sys
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Mapping, Sequence, Type, cast

from rich.table import Table
from rich.text import Text
from yaralyzer.util.helpers.interaction_helper import ask_to_proceed

from epstein_files.documents.config.doc_cfg import Metadata
from epstein_files.documents.config.manual_config import create_configs
from epstein_files.documents.document import Document
from epstein_files.documents.documents.search_result import SearchResult
from epstein_files.documents.doj_file import DojFile
from epstein_files.documents.email import Email
from epstein_files.documents.emails.constants import UNINTERESTING_EMAILERS
from epstein_files.documents.emails.dropsite_email import DropsiteEmail
from epstein_files.documents.emails.util import split_up_leon_black, split_up_dilorio
from epstein_files.documents.json_file import JsonFile
from epstein_files.documents.messenger_log import MSG_REGEX, MessengerLog
from epstein_files.documents.messenger_log_pdf import IMESSAGE_PDF_IDS, MessengerLogPdf
from epstein_files.documents.other_file import OtherFile
from epstein_files.output.rich import TABLE_TITLE_STYLE, console
from epstein_files.people.person import INVALID_FOR_EPSTEIN_WEB, PEOPLE_BIOS, Person
from epstein_files.util.constant.strings import *
from epstein_files.util.constants import *
from epstein_files.util.env import args, logger
from epstein_files.util.helpers.data_helpers import flatten, json_safe, sort_dict_by_keys, uniquify
from epstein_files.util.helpers.file_helper import all_txt_paths, doj_txt_paths, extract_file_id, file_size_str
from epstein_files.util.timer import Timer

# Lists of properties to copy into duplicate documents (will be preceded with 'extracted_')
PROPS_TO_COPY = ['author', 'timestamp']
EMAIL_PROPS_TO_COPY = ['recipients']
SLOW_FILE_SECONDS = 1.0


@dataclass
class EpsteinFiles:
    """
    Attributes:
        file_paths (list[Path]): paths to Epstein related text documents
        documents (list[Document]): all parsed Documents except the emails with was_split_up flag
        timer (Timer): for logging only
        uninteresting_ccs (list[Name]): names of tangential people who were just CCed once or similar
    """
    file_paths: list[Path] = field(init=False)

    # Derived fields
    _documents: list[Document] = field(default_factory=list)
    timer: Timer = field(default_factory=lambda: Timer())
    uninteresting_ccs: list[Name] = field(default_factory=list)
    _empty_file_ids: set[str] = field(default_factory=set)
    _emailers: list[Person] = field(default_factory=list)

    def __post_init__(self):
        """Iterate through files and build appropriate objects."""
        self.file_paths = sorted(all_txt_paths(), reverse=True)
        self._documents = self._load_file_paths(self.file_paths)
        self._documents += split_up_dilorio(self.emails_by(CHRISTOPHER_DILORIO))
        self._documents += split_up_leon_black(self.get_id(LEON_BLACK_EMAIL_ID))
        self._finalize_data_and_write_to_disk()

    @classmethod
    def get_files(cls, timer: Timer | None = None) -> 'EpsteinFiles':
        """Alternate constructor that reads/writes a pickled version of the data ('timer' arg is for logging)."""
        timer = timer or Timer()

        if args.pickle_path.exists() and not args.overwrite_pickle:
            with gzip.open(args.pickle_path, 'rb') as file:
                epstein_files = pickle.load(file)
                timer_msg = f"Loaded {len(epstein_files.documents):,} documents from '{args.pickle_path}'"
                timer.print_at_checkpoint(f"{timer_msg} ({file_size_str(args.pickle_path)})")

            if args.load_new:
                epstein_files.load_new_files()
            elif args.reload_doj:
                epstein_files.reload_doj_files()

            return epstein_files

        logger.warning(f"Building new cache file, this will take a few minutes...")
        epstein_files = EpsteinFiles()
        num_synthetic_cfgs = len([c for c in CONFIGS_BY_ID.values() if c.is_synthetic])
        timer.print_at_checkpoint(f'Processed {len(epstein_files.file_paths):,} files, {num_synthetic_cfgs} synthetic configs')
        return epstein_files

    @property
    def all_doj_files(self) -> Sequence[DojFile | Email]:
        """All files with the filename EFTAXXXXXX, including those that were turned into `Email` objs."""
        return [d for d in self.documents if d.file_info.is_doj_file]

    @property
    def documents(self) -> Sequence[Document]:
        return [d for d in self._documents if not (isinstance(d, Email) and d._was_split_up)]

    @property
    def doj_files(self) -> list[DojFile]:
        """Only returns DojFile type. Emails derived from DOJ files are not included."""
        return [f for f in self.other_files if isinstance(f, DojFile)]

    @property
    def dropsite_emails(self) -> list[DropsiteEmail]:
        return [f for f in self.documents if isinstance(f, DropsiteEmail)]

    @property
    def emails(self) -> list[Email]:
        return [d for d in self.documents if isinstance(d, Email)]

    @property
    def emails_with_attachments(self) -> list[Email]:
        return [e for e in self.emails if e.attached_docs]

    @property
    def emailers(self) -> list[Person]:
        """All the people who sent or received an email."""
        return self._emailers

    @property
    def counterparties_dict(self) -> dict[Name, list[Name]]:
        return sort_dict_by_keys({p.name: p.counterparties for p in self.emailers})

    @property
    def imessage_logs(self) -> list[MessengerLog]:
        return [d for d in self.documents if isinstance(d, MessengerLog)]

    @property
    def json_files(self) -> list[JsonFile]:
        """JSON files from the November document dump, mostly Apple ads related."""
        return [d for d in self.other_files if isinstance(d, JsonFile)]

    @property
    def interesting_other_files(self) -> Sequence[OtherFile]:
        """`OtherFile` objects that have been deemed of interest."""
        return [f for f in self.other_files if f.is_interesting]

    @property
    def local_extracts(self) -> Sequence[Document]:
        """Returns documents that are locally derived from source files."""
        return [d for d in self.documents if d.file_info.is_local_extract_file]

    @property
    def non_duplicate_docs(self) -> Sequence[Document]:
        return Document.without_dupes(self.documents)

    @property
    def non_json_other_files(self) -> list[OtherFile]:
        return [doc for doc in self.other_files if not isinstance(doc, JsonFile)]

    @property
    def other_files(self) -> Sequence[OtherFile]:
        return [d for d in self.documents if isinstance(d, OtherFile)]

    @property
    def uninteresting_emailers(self) -> list[Name]:
        """Emailers whom we don't want to print a separate section for because they're just CCed."""
        if '_uninteresting_emailers' not in vars(self):
            self._uninteresting_emailers = sorted(uniquify(UNINTERESTING_EMAILERS + self.uninteresting_ccs))

        return self._uninteresting_emailers

    @property
    def unique_documents(self) -> Sequence[Document]:
        return [d for d in self.non_duplicate_docs if not d.is_attachment]

    @property
    def unique_emails(self) -> list[Email]:
        """All `Email` objects except for duplicates."""
        return Document.without_dupes(self.emails)

    def docs_matching(self, pattern: re.Pattern | str, names: list[Name] | None = None) -> list[SearchResult]:
        """Find documents whose text matches `pattern` optionally limited to only docs involving `name`)."""
        documents = [d for d in self._documents if (not names) or d.author in names]
        results: list[SearchResult] = []

        for doc in documents:
            if doc.is_duplicate:
                continue

            lines = doc.lines_matching(pattern)

            if args.min_line_length:
                lines = [line for line in lines if len(line.line) > args.min_line_length]

            if len(lines) > 0:
                results.append(SearchResult(doc, lines))

        return results

    def earliest_email_at(self, name: Name) -> datetime:
        """First email timestamp sent to or received by `name`."""
        return self.emails_for(name)[0].timestamp

    def last_email_at(self, name: Name) -> datetime:
        """Last email timestamp sent to or received by `name`."""
        return self.emails_for(name)[-1].timestamp

    def email_author_counts(self) -> dict[Name, int]:
        """Returns dict counting up how many emails were written by each person."""
        return {
            person.name: len(person.unique_emails_by)
            for person in self.emailers if len(person.unique_emails_by) > 0
        }

    def email_authors_to_device_signatures(self) -> dict[str, set[str]]:
        """Mapping of authors to all the device signatures identified in their history."""
        signatures = defaultdict(set)

        for email in [e for e in self.unique_emails if e.sent_from_device]:
            signatures[email.author_or_unknown].add(email.sent_from_device)

        return signatures

    def email_device_signatures_to_authors(self) -> dict[str, set[str]]:
        """Mapping of device signatures to all the users who ever signed an email with them."""
        signatures = defaultdict(set)

        for email in [e for e in self.unique_emails if e.sent_from_device]:
            signatures[email.sent_from_device].add(email.author_or_unknown)

        return signatures

    def email_recipient_counts(self) -> dict[Name, int]:
        """Returns dict counting up how many emails were received by each person."""
        return {p.name: len(p.unique_emails_to) for p in self.emailers if len(p.unique_emails_to) > 0}

    def email_signature_substitution_counts(self) -> dict[str, int]:
        """Return the number of times an email signature was replaced with "<...snipped...>" for each author."""
        substitution_counts = defaultdict(int)

        for email in self.emails:
            for name, num_replaced in email.signature_substitution_counts.items():
                substitution_counts[name] += num_replaced

        return substitution_counts

    def emails_by(self, author: Name) -> list[Email]:
        """All emails sent by `author` (including dupes)."""
        return Document.sort_by_timestamp([e for e in self.emails if e.author == author])

    def emails_for(self, name: Name) -> list[Email]:
        """All emails to or from a given 'name' sorted chronologically (including dupes)."""
        emails = self.emails_by(name) + self.emails_to(name)
        emails += [e for e in self.emails if (e.config and name == e.config.show_with_name)]

        if len(emails) == 0:
            logger.warning(f"No emails found for '{name}'")

        return Document.sort_by_timestamp(Document.uniquify(emails))

    def emails_to(self, name: Name) -> list[Email]:
        """All `Email`s sent to `name` (including dupes)."""
        if name is None:
            emails = [e for e in self.emails if len(e.recipients) == 0 or None in e.recipients]
        else:
            emails = [e for e in self.emails if name in e.recipients]

        return Document.sort_by_timestamp(emails)

    def get_ids(self, ids: list[str], rebuild: bool = False) -> Sequence[Document]:
        """Get `Document` objects for `file_ids`. If `rebuild` is True then rebuild `Document` from .txt file."""
        docs = [d for d in self._documents if d.file_id in ids]

        if len(docs) != len(ids):
            logger.warning(f"{len(ids)} file IDs provided but only {len(docs)} documents found! ids: {ids}")

        return [d.reload() if rebuild else d for d in docs]

    def get_id(self, file_id: str, rebuild: bool = False, required_type: Type[Document] = Document) -> Document:
        """Singular ID version of `get_ids()` but with option to require a type of document subclass."""
        doc = self.get_ids([file_id], rebuild)[0]

        if required_type:
            if not isinstance(doc, required_type):
                raise ValueError(f"No {required_type.__name__} found for {file_id} (found {doc})")

        return doc

    def imessage_logs_for(self, name: Name) -> list[MessengerLog]:
        """Return `MessengerLog` objects where Epstein's counterparty is `name`."""
        return [log for log in self.imessage_logs if name == log.author]

    def json_metadata(self) -> str:
        """Create a JSON string containing metadata for all the files."""
        metadata = {
            'files': {
                Email.__name__: _sorted_metadata(self.emails),
                JsonFile.__name__: _sorted_metadata(self.json_files),
                MessengerLog.__name__: _sorted_metadata(self.imessage_logs),
                OtherFile.__name__: _sorted_metadata(self.non_json_other_files),
            },
            'people': {name: bio.plain for name, bio in PEOPLE_BIOS.items()},
        }

        return json.dumps(metadata, indent=4, sort_keys=True)

    def person_objs(self, names: list[Name]) -> list[Person]:
        """Construct Person objects for a list of names."""
        return [
            Person(
                name=name,
                emails=self.emails_for(name),
                imessage_logs=self.imessage_logs_for(name),
                is_uninteresting=name in self.uninteresting_emailers,
                other_files=[
                    f for f in self.other_files
                    if name and (name == f.author or (f.config and name == f.config.show_with_name))
                ]
            )
            for name in names
        ]

    def load_new_files(self) -> None:
        """Load any new files detected in the hierarchy."""
        new_docs = self._load_file_paths(self._new_files())

        if not new_docs:
            logger.warning(f"No new files found, doing nothing.")
            return
        elif (cfgs := create_configs(new_docs)):  # Exit if new configs were created
            console.print(f"\n{len(cfgs)} config objects created, exiting...", style='bold yellow1')
            sys.exit()

        self._finalize_data_and_write_to_disk(new_docs)

    def reload_doj_files(self) -> None:
        """Reload only the DOJ PDF extracts (keep HOUSE_OVERSIGHT stuff unchanged)."""
        def doj_file_counts_str():
            return f"(have {len(self.all_doj_files)}, {len(self.doj_files)} non-email)"

        timer = Timer()
        logger.warning(f"Reloading all DOJ files {doj_file_counts_str()}...")
        self._finalize_data_and_write_to_disk(self._load_file_paths(doj_txt_paths()))
        timer.print_at_checkpoint(f"Reloaded {len(self.doj_files)} DOJ files {doj_file_counts_str()}")

    def overview_table(self) -> Table:
        """Table showing file counts by type."""
        title = Text('Files Overview ', TABLE_TITLE_STYLE)
        title.append('(last updated: ', 'dim').append(datetime.now().date().isoformat(), 'cyan').append(')', 'dim')
        table = Document.files_summary_table(title, 'File Type')
        table.add_row('Emails', *Document.file_summary_row(self.emails))
        table.add_row('iMessage Logs', *Document.file_summary_row(self.imessage_logs))
        table.add_row('JSON Data', *Document.file_summary_row(self.json_files, True))
        table.add_row('Other', *Document.file_summary_row(self.non_json_other_files))
        return table

    def repair_ids(self, ids: list[str]) -> None:
        """Reload the `ids` and save updated pickle file (also loads new files)."""
        ids = uniquify(ids)

        if ids == ['EMAIL']:
            doc_paths = [d.file_path for d in self.emails]
            msg = f"Repairing all {len(doc_paths)} emails..."
        else:
            doc_paths = [d.file_path for d in self._documents if d.file_id in ids]
            msg = f"Repairing {len(ids)} file IDs"

            if len(doc_paths) != len(ids):
                raise RuntimeError(f"{len(ids)} specified but only {len(doc_paths)} Document objects found!")

        if (new_paths := self._new_files()):
            msg +=  f" (also loading {len(new_paths)} new files)"
            doc_paths += new_paths

        logger.warning(msg)
        self._finalize_new_docs_if_approved(self._load_file_paths(doc_paths))

    def save_to_disk(self) -> None:
        """Write a pickled version of this `EpsteinFiles` object with all documents etc."""
        with gzip.open(args.pickle_path, 'wb') as file:
            pickle.dump(self, file)
            logger.warning(f"Pickled data to '{args.pickle_path}' ({file_size_str(args.pickle_path)})...")

    def unknown_recipient_ids(self) -> list[str]:
        """IDs of emails whose recipient is not known."""
        return sorted([e.file_id for e in self.emails if None in e.recipients or not e.recipients])

    def _copy_duplicate_doc_properties(self) -> None:
        """Ensure dupe emails have the properties of the emails they duplicate to capture any repairs, config etc."""
        for doc in self.documents:
            if not doc.duplicate_of_id:
                continue

            original = self.get_id(doc.duplicate_of_id)
            props_to_copy = (EMAIL_PROPS_TO_COPY + PROPS_TO_COPY) if isinstance(doc, Email) else PROPS_TO_COPY
            props_to_copy = [f"extracted_{prop}" for prop in props_to_copy]

            for field_name in props_to_copy:
                original_prop = getattr(original, field_name)
                duplicate_prop = getattr(doc, field_name)

                if original_prop != duplicate_prop:
                    doc.warn(f"Replacing {field_name} {duplicate_prop} with {original_prop} from duplicated '{original.file_id}'")
                    setattr(doc, field_name, original_prop)

    def _docs_by_id(self) -> Mapping[str, Document]:
        return {doc.file_id: doc for doc in self._documents}

    def _finalize_data_and_write_to_disk(self, new_docs: list[Document] | None = None) -> None:
        """Handle computation of fields related to uninterestingness, relationships between documents, etc."""
        new_docs = new_docs or []
        # dilorio_files = [e.reload() for e in self._documents if e.author == CHRISTOPHER_DILORIO and e.file_info.has_file]
        # new_docs += dilorio_files + split_up_dilorio(dilorio_files)

        if new_docs:
            old_num_docs = len(self._documents)
            new_doc_ids = [d.file_info.file_id for d in new_docs]
            self._documents = [d for d in self._documents if d.file_info.file_id not in new_doc_ids]  # Remove existing
            logger.warning(f"Adding {len(new_docs)} Documents (replacing {old_num_docs - len(self._documents)} existing)")
            self._documents += new_docs

        self._set_uninteresting_ccs()
        self._copy_duplicate_doc_properties()
        self._emailers = self.person_objs(flatten([e.participants for e in self.emails]))
        self._find_email_attachments_and_set_is_first_for_user()
        self._documents = Document.sort_by_timestamp(self._documents)
        self.save_to_disk()

    def _find_email_attachments_and_set_is_first_for_user(self) -> None:
        email_attachments = [f for f in self.other_files if f.config and f.config.attached_to_email_id]

        for email in self.emails:
            email.attached_docs = []  # Remove all attachments before re-finding them if it's a repair

        for file in email_attachments:
            email: Email = self.get_id(file.config.attached_to_email_id, required_type=Email)

            if email.is_duplicate:
                raise ValueError(f"Cannot attach {file.file_id} to duplicate email {email}")

            email.attached_docs.append(file)

            if file.config.timestamp:  # Don't overwrite configured timestamps (think of a book or article attachment)
                continue
            elif file.timestamp and file.timestamp != email.timestamp:
                file.warn(f"Overwriting '{file.timestamp}' with {email}'s timestamp {email.timestamp}")

            file.extracted_timestamp = email.timestamp

        # Set the is_persons_first_email flag on the earliest Email we have for each person.
        for emailer in self.emailers:
            if emailer.name in INVALID_FOR_EPSTEIN_WEB or len(emailer.unique_emails) == 0:
                continue

            emailer.unique_emails[0].is_persons_first_email = True

    def _finalize_new_docs_if_approved(self, new_docs: list[Document]) -> None:
        """Same as _finalize_data_and_write_to_disk() but prints new docs and asks for permission."""
        if len(new_docs) < 100:
            console.print(*new_docs)
        else:
            logger.warning(f"Not showing the {len(new_docs)} repaired documents...")

        logger.warning(f"Finalizing {len(new_docs)} files: {[d.file_id for d in new_docs]}")
        ask_to_proceed("Looks good?")
        self._finalize_data_and_write_to_disk(new_docs)

    def _load_file_paths(self, file_paths: list[Path]) -> list[Document]:
        """Load a list of file paths into a list of `Document` object subclasses."""
        docs: list[Document] = []

        for file_path in file_paths:
            doc_timer = Timer(decimals=2)
            document = Document(file_path)

            if document.length == 0:
                if document.file_id not in self._empty_file_ids:
                    document.warn(f"Skipping empty file...")
                    self._empty_file_ids.add(document.file_id)

                continue

            cls = document_cls(document)
            # docs.append(cls(file_path, lines=document.lines, text=document.text))
            docs.append(cls(file_path))  # TODO: needs to reload DropsiteEmail
            logger.info(str(docs[-1]))

            if doc_timer.seconds_since_start() > SLOW_FILE_SECONDS:
                doc_timer.print_at_checkpoint(f"Slow file: {docs[-1]} processed")

        return docs

    def _new_files(self) -> list[Path]:
        """Find any files that don't exist in the collection. Has side effect of setting `self.file_paths`."""
        self.file_paths = sorted(all_txt_paths(), reverse=True)
        current_doc_ids = [id for id in self._docs_by_id().keys()] + list(self._empty_file_ids)

        return [
            p for p in self.file_paths
            if extract_file_id(p) not in current_doc_ids
        ]

    def _set_uninteresting_ccs(self) -> None:
        """Extract the recipients of emails configured has having uninteresting CCs or BCCs."""
        for email in [e for e in self.emails if e.config and e.config.has_uninteresting_bccs]:
            self.uninteresting_ccs += [bcc.lower() for bcc in cast(list[str], email.header.bcc)]

        for email in [e for e in self.emails if e.config and e.config.has_uninteresting_ccs]:
            self.uninteresting_ccs += email.recipients

        self.uninteresting_ccs = sorted(uniquify(self.uninteresting_ccs))
        logger.info(f"Extracted uninteresting_ccs: {self.uninteresting_ccs}")


def document_cls(doc: Document) -> Type[Document]:
    """Find the appropriate `Document` subclass for this file based on the contents."""
    if doc.length == 0:
        return Document
    elif doc.file_info.is_eml_file:
        return DropsiteEmail
    elif doc.text[0] == '{':
        return JsonFile
    elif doc.file_id in IMESSAGE_PDF_IDS:
        return MessengerLogPdf
    elif doc.is_email:
        return Email
    elif doc.file_info.is_doj_file:
        return DojFile
    elif MSG_REGEX.search(doc.text[0:5000]):  # Limit search area to avoid pointless scans of huge files
        return MessengerLog
    else:
        return OtherFile


def _sorted_metadata(docs: Sequence[Document]) -> list[Metadata]:
    return [json_safe(d.metadata) for d in Document.sort_by_id(docs)]



from epstein_files.documents.config.categories.legal import (EDWARDS_V_DERSHOWITZ, JANE_DOE_V_EPSTEIN_TRUMP,
     JANE_DOE_2_V_EPSTEIN, EPSTEIN_V_ROTHSTEIN_EDWARDS, REDACTED_V_EPSTEIN_ESATE)
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *


DEPOSITION_CFGS = [
    DocCfg(id='021824', author=PAUL_G_CASSELL, description=f"from {EDWARDS_V_DERSHOWITZ}"),
    DocCfg(id='013463', author='Scott Rothstein', description=f"from {JANE_DOE_V_EPSTEIN_TRUMP}", date='2010-03-23'),
    DocCfg(id='017488', author='Scott Rothstein', description=f"from {EPSTEIN_V_ROTHSTEIN_EDWARDS}", date='2012-06-22'),
    # DOJ
    DocCfg(id='EFTA00615804', author=ALAN_DERSHOWITZ, date='2016-01-12', description='pages 334-461 (heavily redacted)'),
    DocCfg(
        id='EFTA00159483',
        author=LAWRANCE_VISOSKI,
        is_interesting=True,
        non_participants=[EHUD_BARAK, GLENN_DUBIN, LARRY_SUMMERS, NAOMI_CAMPBELL],
    ),
    DocCfg(id='EFTA00009229', author='Alex Acosta', description='pages 1-100', date='2020-04-30', is_interesting=True),
    DocCfg(id='EFTA00009329', author='Alex Acosta', description='pages 101-200', date='2020-04-30', is_interesting=True),
    DocCfg(id='EFTA00009016', author='Alex Acosta', description='pages 201-300', date='2020-04-30', is_interesting=True),
    DocCfg(id='EFTA00009116', author='Alex Acosta', description='pages 300-411', date='2020-04-30', is_interesting=True),
    DocCfg(
        id='EFTA00008744',
        author='FBI',
        date='2021-03-29',
        description='grand jury testimony of Child Exploitation & Human Trafficking Task Force member',
    ),
    DocCfg(id='EFTA01108807', author='Jane Doe', date='2010-02-09', description='vol. III of IV'),
    DocCfg(id='EFTA00023557', author='victim', description=f'who says {GHISLAINE_MAXWELL} and Epstein pay for nudes'),
    # DocCfg(id='EFTA00064843', author=UNKNOWN_GIRL, description=f'{GHISLAINE_MAXWELL} "relentless"'),
    DocCfg(id='EFTA00078311', author=REDACTED_V_EPSTEIN_ESATE, description=f'Boies Schiller filing about sexual abuse of their client'),
    DocCfg(id='EFTA00105975', author=REDACTED_V_EPSTEIN_ESATE, description=f'{GHISLAINE_MAXWELL} "was relentless"'),
    DocCfg(id='EFTA00158752', author=REDACTED_V_EPSTEIN_ESATE, description=f'{GHISLAINE_MAXWELL} "was relentless"'),
    DocCfg(id='EFTA01246559', author=JANE_DOE_2_V_EPSTEIN, date='2010-04-09'),
]



"""
Small methods for building rich text constructs.
"""
from rich.padding import Padding
from rich.text import Text

left_indent_padding = lambda num_spaces: (0, 0, 0, num_spaces)
left_indent = lambda obj, num_spaces: Padding(obj, left_indent_padding(num_spaces))
no_bold = lambda style: style.replace('bold', '').strip()
vertically_pad = lambda obj, amount = 1: Padding(obj, (amount, 0, amount, 0))


def enclose(txt: str | Text, encloser: str = '', encloser_style: str = 'wheat4') -> Text:
    """Surround with something."""
    if not encloser:
        return Text('').append(txt)
    elif len(encloser) != 2:
        raise ValueError(f"'encloser' arg is '{encloser}' which is not 2 characters long")

    enclose_start, enclose_end = (encloser[0], encloser[1])
    return Text('').append(enclose_start, encloser_style).append(txt).append(enclose_end, encloser_style)


def join_non_empty(*txts, sep: str | Text = ' ') -> Text:
    """Join two strings but only if they are not empty."""
    sep = sep if isinstance(sep, Text) else Text(sep)
    return sep.join([t for t in txts if len(t) > 0])



import re
from dataclasses import dataclass, field
from datetime import datetime
from typing import cast

from rich.text import Text

from epstein_files.documents.document import CLOSE_PROPERTIES_CHAR, Document
from epstein_files.documents.config.doc_cfg import CommunicationCfg
from epstein_files.output.highlight_config import HIGHLIGHTED_CONTACTS, get_style_for_name, styled_name
from epstein_files.output.rich import styled_key_value
from epstein_files.people.names import UNKNOWN, Name
from epstein_files.util.helpers.data_helpers import uniquify
from epstein_files.util.helpers.rich_helpers import no_bold

TIMESTAMP_SECONDS_REGEX = re.compile(r":\d{2}$")


@dataclass
class Communication(Document):
    """
    Superclass for `Email` and `MessengerLog`.

    Attributes:
        recipients (list[Name]): People to whom this email was sent.
    """
    extracted_recipients: list[Name] = field(default_factory=list)

    def __post_init__(self):
        super().__post_init__()
        self.extracted_recipients = [] if self.config and self.config.recipients else self._extract_recipients()

    @property
    def author_or_unknown(self) -> str:
        return self.author or UNKNOWN

    @property
    def author_style(self) -> str:
        return get_style_for_name(self.author)

    @property
    def author_txt(self) -> Text:
        return styled_name(self.author)

    @property
    def border_style(self) -> str:
        return no_bold(self.author_style)

    @property
    def people(self) -> list[str]:
        """Names of people who either sent/received this email or are mentioned in it."""
        if self.config and self.config.people:
            return self.config.people

        people = super().people + [p for p in self.participants if p]
        return sorted(uniquify(people))

    @property
    def config(self) -> CommunicationCfg | None:
        """Configured timestamp, if any."""
        cfg = super().config

        if cfg and not isinstance(cfg, CommunicationCfg):
            self.warn(f"Found config that's the wrong type! {repr(cfg)}")
            cfg = cast(CommunicationCfg, cfg)

        return cfg

    @property
    def has_unknown_participant(self) -> bool:
        return None in self.participants

    @property
    def has_unknown_recipient(self) -> bool:
        return None in (self.recipients or [None])

    @property
    def is_recipient_uncertain(self) -> bool:
        return bool(self.config and self.config.recipient_uncertain)

    @property
    def participants(self) -> set[Name]:
        """Author + recipients including a `None` if `self.recipients` is empty."""
        return set([self.author] + (self.recipients or [None]))

    @property
    def recipients(self) -> list[Name]:
        if self.config and self.config.recipients:
            return self.config.recipients
        else:
            return self.extracted_recipients

    @property
    def recipient_style(self) -> str:
        return get_style_for_name((self.recipients or [self.author])[0])

    @property
    def timestamp_without_seconds(self) -> str:
        return TIMESTAMP_SECONDS_REGEX.sub('', str(self.timestamp))

    @property
    def _summary(self) -> Text:
        """One line summary mostly for logging."""
        return self._summary_with_author.append(CLOSE_PROPERTIES_CHAR)

    @property
    def _summary_with_author(self) -> Text:
        """Append author information to `super().summary`, bracket is left open."""
        author_str = styled_key_value('author', Text(f"'{self.author_or_unknown}'", style=self.author_style))
        return super()._summary.append(', ').append(author_str)

    def colored_external_links(self) -> Text:
        """Overrides super() method to apply `self.author_style`."""
        return self.file_info.build_external_links(self.recipient_style, with_alt_links=True)

    def _extract_recipients(self) -> list[Name]:
        """Overload in subclasses"""
        return []

    @classmethod
    def default_category(cls) -> str:
        return cls.__name__.lower()



from epstein_files.documents.config.doc_cfg import DocCfg, EmailCfg
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *

AL_SECKEL_BILL_FIGHT = f"{AL_SECKEL} and Epstein fight about the bill for reputation management services"


REPUTATION_CFGS = [
    DocCfg(id='030426', author=IAN_OSBORNE, description=f"reputation repair proposal citing Michael Milken", date='2011-06-14'),
    DocCfg(id='026582', description=f"Epstein's internet search results at start of reputation repair campaign, maybe from {OSBORNE_LLP}"),
    DocCfg(id='030573', description=f"Epstein's unflattering Google search results, maybe screenshot by {AL_SECKEL} or {OSBORNE_LLP}"),
    DocCfg(id='030875', description=f"Epstein's Wikipedia page", date='2014-02-08'),  # Date is based on tyler shears; seckel was 2010
    DocCfg(id='026583', description=f"Google search results for '{JEFFREY_EPSTEIN}' with analysis ({OSBORNE_LLP}?)"),
    DocCfg(id='029350', description=f"Microsoft Bing search results for Epstein with sex offender at top", attached_to_email_id='EFTA00675542'),
    EmailCfg(id='022203', description=AL_SECKEL_BILL_FIGHT, truncate_to=500),
    EmailCfg(id='022219', description=AL_SECKEL_BILL_FIGHT, truncate_to=2404),

    # DOJ files
    EmailCfg(id='EFTA01830035', description=AL_SECKEL_BILL_FIGHT),
    DocCfg(
        id='EFTA01810372',
        author=TYLER_SHEARS,
        description=f'invoice for reputation management work',
        is_interesting=True,
        attached_to_email_id='EFTA01931256'
    ),
]



"""
Custom configurations for iMessage logs.
"""
from epstein_files.documents.documents.categories import Interesting
from epstein_files.documents.config.doc_cfg import TextCfg
from epstein_files.people.names import *

PARTICIPANTS_FIELD = 'Participants: field'


################################################################################################
############################################ TEXTS #############################################
################################################################################################

CONFIRMED_TEXTS_CONFIG = [
    TextCfg(id='031042', author=ANIL_AMBANI, author_reason=PARTICIPANTS_FIELD),
    TextCfg(id='027225', author=ANIL_AMBANI, author_reason="birthday mentioned in texts is confirmed as Ambani's"),
    TextCfg(id='031054', author=ANTHONY_SCARAMUCCI, author_reason="Scaramucci's phone number is at top of raw file"),
    TextCfg(id='027333', author=ANTHONY_SCARAMUCCI, author_reason="Scaramucci's phone number is in one of the messages"),
    TextCfg(id='031173', author=ARDA_BESKARDES, author_reason=PARTICIPANTS_FIELD),
    TextCfg(id='027401', author=EVA, author_reason=PARTICIPANTS_FIELD),
    TextCfg(id='027650', author=JOI_ITO, author_reason=PARTICIPANTS_FIELD),
    TextCfg(id='027777', author=LARRY_SUMMERS, author_reason=PARTICIPANTS_FIELD),
    TextCfg(id='027515', author=MIROSLAV_LAJCAK, author_reason='https://x.com/ImDrinknWyn/status/1990210266114789713'),
    TextCfg(
        id='027165',
        author=MELANIE_WALKER,
        author_reason='says "it\'s Melanie", also https://www.wired.com/story/jeffrey-epstein-claimed-intimate-knowledge-of-donald-trumps-views-in-texts-with-bill-gates-adviser/',
        category=Interesting.CRYPTO,
        comment='crypto health'
    ),
    TextCfg(id='027248', author=MELANIE_WALKER, author_reason='says "we met through Trump" which is confirmed by Melanie in 032803'),
    TextCfg(id='025429', author=STACEY_PLASKETT, author_reason='widely reported'),
    TextCfg(id='027128', author=SOON_YI_PREVIN, author_reason='https://x.com/ImDrinknWyn/status/1990227281101434923'),
    TextCfg(id='027217', author=SOON_YI_PREVIN, author_reason='refs marriage to Woody Allen'),
    TextCfg(id='027244', author=SOON_YI_PREVIN, author_reason='refs Woody Allen'),
    TextCfg(id='027257', author=SOON_YI_PREVIN, author_reason=f"Woody Allen in {PARTICIPANTS_FIELD}"),
    TextCfg(
        id='027460',
        author=STEVE_BANNON,
        author_reason='Discusses leaving scotland when Bannon was confirmed in Scotland, also NYT',
        is_interesting=True,
        comment='bannon and jabor',
    ),
    TextCfg(id='027307', author=STEVE_BANNON, author_reason='texts mention "Epstein Bannon Kurz"'),
    TextCfg(id='027278', author=TERJE_ROD_LARSEN, author_reason=PARTICIPANTS_FIELD, is_interesting=True),
    TextCfg(id='027255', author=TERJE_ROD_LARSEN, author_reason=PARTICIPANTS_FIELD),
]

UNCONFIRMED_TEXTS_CONFIG = [
    TextCfg(id='027762', author=ANDRZEJ_DUDA, author_uncertain=f"Duda in NY at that time, took train"),
    TextCfg(id='027774', author=ANDRZEJ_DUDA, author_uncertain=f"Duda in NY at that time, took train"),
    TextCfg(id='027221', author=ANIL_AMBANI),
    TextCfg(id='027396', author=ANTHONY_SCARAMUCCI, author_uncertain='says "I need to make peace with Bannon"'),
    TextCfg(id='025436', author=CELINA_DUBIN),
    TextCfg(id='027576', author=MELANIE_WALKER, author_uncertain='https://www.ahajournals.org/doi/full/10.1161/STROKEAHA.118.023700'),
    TextCfg(id='027141', author=MELANIE_WALKER),
    TextCfg(id='027232', author=MELANIE_WALKER),
    TextCfg(id='027133', author=MELANIE_WALKER),
    TextCfg(id='027184', author=MELANIE_WALKER),
    TextCfg(id='027214', author=MELANIE_WALKER),
    TextCfg(id='027148', author=MELANIE_WALKER),
    TextCfg(id='027440', author=MICHAEL_WOLFF, author_uncertain='AI says Trump book/journalism project'),
    TextCfg(id='027568', author=STEVE_BANNON),
    TextCfg(id='027695', author=STEVE_BANNON, description='"kazak daughter the key" may be about Dariga Nazarbayeva (low confidence)'),
    TextCfg(id='027594', author=STEVE_BANNON),
    TextCfg(id='027549', author=STEVE_BANNON),
    TextCfg(id='027764', author=STEVE_BANNON, description=f"includes discussion of Facebook's attempt at crypto called Libra", is_interesting=True),
    TextCfg(id='027585', author=STEVE_BANNON, author_uncertain='references Tokyo trip'),
    TextCfg(id='027434', author=STEVE_BANNON, author_uncertain='references Maher appearance'),
    TextCfg(id='027428', author=STEVE_BANNON, author_uncertain='references HBJ meeting on 9/28 from other Bannon/Epstein convo'),
    TextCfg(id='027374', author=STEVE_BANNON, author_uncertain='AI says China strategy and geopolitics'),
    TextCfg(id='027455', author=STEVE_BANNON, author_uncertain='AI says China strategy and geopolitics; Trump discussions'),
    TextCfg(id='027536', author=STEVE_BANNON, author_uncertain='AI says China strategy and geopolitics; Trump discussions'),
    TextCfg(id='025479', author=STEVE_BANNON, author_uncertain='AI says China strategy and geopolitics; Trump discussions'),
    TextCfg(id='025734', author=STEVE_BANNON, author_uncertain='AI says China strategy and geopolitics; Trump discussions'),
    TextCfg(id='027707', author=STEVE_BANNON, author_uncertain='AI says Italian politics; Trump discussions'),
    TextCfg(id='025400', author=STEVE_BANNON, author_uncertain='AI says Trump NYT article criticism; Hannity media strategy'),
    TextCfg(id='025408', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage'),
    TextCfg(id='025452', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage'),
    TextCfg(id='027281', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage'),
    TextCfg(id='027346', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage'),
    TextCfg(id='027365', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage'),
    TextCfg(id='027406', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage'),
    TextCfg(id='027655', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage'),
    TextCfg(id='027722', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage', duplicate_ids=['027720']),
    TextCfg(id='027735', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage', description=f"Bannon just returned from Kazakhstan"),
    TextCfg(id='027794', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage'),
    TextCfg(id='025368', author=STEVE_BANNON, author_uncertain='AI', is_interesting=True, comment='obama slander'),
    TextCfg(id='025707', author=STEVE_BANNON, author_uncertain='AI', is_interesting=True, comment='mentions thiel, dubai'),
    TextCfg(id='025363', author=STEVE_BANNON, author_uncertain='AI', is_interesting=True, comment='mooch discussion'),
    TextCfg(id='027445', author=STEVE_BANNON, author_uncertain='AI', is_interesting=True),
    TextCfg(id='027260', author=STEVE_BANNON, author_uncertain='AI', is_interesting=True),
    TextCfg(
        id='029744',
        author=STEVE_BANNON,
        author_uncertain='AI says Trump + NYT coverage',
        is_interesting=True,
        description='contains discussion of Chinese criminal Miles Guo AKA Miles Kwok',
        people=[JEFFREY_EPSTEIN, MILES_GUO, STEVE_BANNON]
    ),
    TextCfg(id='031045', author=STEVE_BANNON, author_uncertain='AI says Trump and New York Times coverage'),
    TextCfg(id='027275', is_interesting=True, comment='"Crypto- Kerry- Qatar -sessions"'),
    TextCfg(id='027792', duplicate_ids=['027694'], dupe_type='quoted'),

    # DOJ PDFs
    TextCfg(
        id='EFTA00508054',
        description=f'Multiple counterparties, at least some Bannon. Epstein brings up the suicide of Ruslana Korshunova whom he "knew very well".'
    ),
]

# Necessary because some of these will be marked uninteresting, which overrides the config
UNKNOWN_TXT_MSG_LOGS = [
    TextCfg(id='025423'),
    TextCfg(id='025426'),
    TextCfg(id='027583'),
]

for cfg in UNCONFIRMED_TEXTS_CONFIG:
    cfg.author_uncertain = cfg.author_uncertain or True

TEXTS_CONFIG = CONFIRMED_TEXTS_CONFIG + UNCONFIRMED_TEXTS_CONFIG + UNKNOWN_TXT_MSG_LOGS



import re
from datetime import datetime

from epstein_files.documents.document import Document
from epstein_files.documents.email import Email
from epstein_files.output.rich import console
from epstein_files.people.names import JEFFREY_EPSTEIN, LEON_BLACK, MELANIE_SPINELLA, sort_names
from epstein_files.util.constant.strings import LEON_BLACK_EMAIL_ID
from epstein_files.util.env import args
from epstein_files.util.logging import logger

DILORIO_SPLIT = '\nFrom: Chris'
LEON_BLACK_EMAIL_REGEX = re.compile(r"^(From: .{,50}\nDate:|Date: ).*?(?=(From|Date|\Z))", re.DOTALL | re.MULTILINE)
LEON_BLACK_FWD_REGEX = re.compile(r"^-+(Forwarded|Original) message-+", re.MULTILINE)
TO_LEON_REGEX = re.compile(r"^Leon,", re.MULTILINE)
TO_JEFFREY_REGEX = re.compile(r"^Jeffrey-", re.MULTILINE)

LEON_BLACK_OCR_REPAIRS = {
    re.compile('^ate: ', re.MULTILINE): 'Date: ',
}


def split_up_dilorio(dilorio_emails: list[Email]) -> list[Email]:
    """Create emails from gigantic Chris Dilorio emails and mark the big email with `_was_split_up`."""
    sub_emails = []
    skipped = []

    for big_email in dilorio_emails:
        texts = [
            (f"{DILORIO_SPLIT}{t}" if i > 0 else t).strip()
            for i, t in enumerate(big_email.text.split(DILORIO_SPLIT))
        ]

        big_email._was_split_up = True
        logger.warning(f"Parsing {big_email} into {len(texts)} sub emails...")

        for i, text in enumerate(texts, 1):
            new_file_stem = _new_file_stem(big_email, i)
            email = Email(big_email.file_path.parent.joinpath(new_file_stem), text=text)
            email.extracted_author = big_email.author
            email.extracted_recipients = email.extracted_recipients or ['SEC']

            if not email.actual_text or email.lines[-1].startswith('Subject:'):
                email.warn(f"skipping empty {email.author} email...")
                skipped.append(email)
                continue

            sub_emails.append(email)

    return _uniquify_by_timestamp(sub_emails, dilorio_emails, 'dilorio', len(skipped))


def split_up_leon_black(big_email: Email) -> list[Email]:
    """Create emails from gigantic Leon Black emails and mark the big email with `_was_split_up`."""
    if big_email.file_id != LEON_BLACK_EMAIL_ID:
        raise ValueError(f"wrong Leon Black email")

    big_text = big_email.repair_ocr_text(LEON_BLACK_OCR_REPAIRS, big_email.text)
    emails: list[Email] = []
    skipped = []
    fwded_email_text = ''

    for i, match in enumerate(LEON_BLACK_EMAIL_REGEX.finditer(big_text), 1):
        text = match.group(0)
        new_file_stem = _new_file_stem(big_email, i)

        if LEON_BLACK_FWD_REGEX.search(text):
            logger.warning(f"Fwded email detected, appending next email...")
            fwded_email_text = fwded_email_text + text
            continue
        else:
            email_text = fwded_email_text + text
            fwded_email_text = ''

        if len(email_text) < 80:
            logger.warning(f"Skipping {new_file_stem} because it's too short...")
            skipped.append(email_text)
            continue

        email = Email(big_email.file_path.parent.joinpath(new_file_stem), text=email_text)
        text_to_scan_for_recipients = email.actual_text[:120]

        if (TO_LEON_REGEX.search(email.text) and LEON_BLACK not in email.extracted_recipients) \
                or email.extracted_recipients == [MELANIE_SPINELLA]:
            email.extracted_recipients = sort_names(email.extracted_recipients + [LEON_BLACK])

        if 'Clapp' in text_to_scan_for_recipients:
            email.extracted_recipients += ['Ada Clapp']

        if 'Avantario' in text_to_scan_for_recipients:
            email.extracted_recipients += ['Joe Avantario']

        if email.recipients and not (email.extracted_author or TO_JEFFREY_REGEX.search(text_to_scan_for_recipients)):
            email.extracted_author = JEFFREY_EPSTEIN

            if JEFFREY_EPSTEIN in email.extracted_recipients:
                email.extracted_recipients = [r for r in email.extracted_recipients if r != JEFFREY_EPSTEIN]

        if args.raw or args.debug:
            print(f"\n--- RAW TEXT {new_file_stem} ---\n{email_text}\n--- END RAW TEXT {new_file_stem} ---\n")
            console.print(email)
            console.print(email._debug_txt())

        emails.append(email)

    big_email._was_split_up = True
    return _uniquify_by_timestamp(emails, [big_email], LEON_BLACK, len(skipped))


def _new_file_stem(email: Email, i: int) -> str:
    return email.file_info.file_stem + f'_{i}.txt'


def _uniquify_by_timestamp(emails: list[Email], from_emails: list[Email], label: str, num_skipped: int = 0) -> list[Email]:
    """Uniquify by timestamp to eliminate dupes."""
    # Validate timestamps
    for email in emails:
        if email.timestamp > datetime(2026, 1, 1):
            raise ValueError(f"{email} bad timestamp in extracted email: {email.timestamp}")

    uniq_emails_by_timestamp = {e.timestamp: e for e in Document.sort_by_id(emails)}

    logger.warning(
        f"Created {len(emails)} ({len(uniq_emails_by_timestamp)} unique) Emails from {len(from_emails)} {label} emails" \
        f", skipped {num_skipped} empty sub emails"
    )

    return [e for e in uniq_emails_by_timestamp.values()]



from epstein_files.documents.config.config_builder import fedex_invoice, important_messages_pad
from epstein_files.documents.config.doc_cfg import DocCfg
from epstein_files.documents.documents.categories import Interesting, Neutral
from epstein_files.documents.doj_files.full_text import EFTA00009622_TEXT
from epstein_files.people.names import *
from epstein_files.util.constant.strings import *

FEMALE_HEALTH_COMPANY = 'Female Health Company (FHC)'
LUTNICKS_CANTOR = f"Howard Lutnick's {CANTOR}"


MISC_CFGS = [
    DocCfg(id='029326', category=Neutral.PRESSER, author=EPSTEIN_FOUNDATION, date='2013-02-15'),
    DocCfg(id='026565', category=Neutral.PRESSER, author=EPSTEIN_FOUNDATION, comment=f'maybe a draft of 029326', date='2013-02-15'),
    DocCfg(id='022494', author='DOJ', description=f'Foreign Corrupt Practices Act (FCPA) Resource Guide'),
    DocCfg(id='023096', author=EPSTEIN_FOUNDATION, description=f'blog post', date='2012-11-15'),
    DocCfg(id='027071', author=FEMALE_HEALTH_COMPANY, description=f"brochure requesting donations for female condoms in Uganda"),
    DocCfg(id='027074', author=FEMALE_HEALTH_COMPANY, description=f"pitch deck (USAID was a customer)"),
    DocCfg(id='032735', author=GORDON_GETTY, description=f"on Trump", date='2018-03-20'),  # Dated based on concurrent emails from Getty
    DocCfg(id='025540', author=JEFFREY_EPSTEIN, description=f"rough draft of his side of the story"),
    DocCfg(id='026634', author='Michael Carrier', description=f'comments about an Apollo linked fund "DE Fund VIII"'),
    DocCfg(id='031425', author=SCOTT_J_LINK, description=f'completely redacted email', is_interesting=False),
    DocCfg(id='020447', author='Working Group on Chinese Influence Activities in the US', description=f'Promoting Constructive Vigilance'),
    DocCfg(id='012718', description=f"{CVRA} congressional record", date='2011-06-17'),
    DocCfg(id='019448', description=f"Haitian business investment proposal called Jacmel", attached_to_email_id='019446'),
    DocCfg(id='023644', description=f"interview with Mohammed bin Salman", date='2016-04-25', is_interesting=False),
    DocCfg(
        id='030142',
        author=f"{KATHRYN_RUEMMLER} & {KEN_STARR}",
        date='2016-09-01',
        description=f"mostly empty {JASTA} (Justice Against Sponsors of Terrorism Act) doc referencing suit against Saudis",
    ),
    DocCfg(
        id='033338',
        category=Neutral.PRESSER,
        date='2000-06-07',
        description=f"end of {DONALD_TRUMP} & {NICHOLAS_RIBIS} working relationship at Trump's casino",
        is_interesting=True,
    ),
    DocCfg(id='029328', description=f"Rafanelli Events promotional deck", is_interesting=False),
    DocCfg(id='029475', description=f'{VIRGIN_ISLANDS} Twin City Mobile Integrated Health Services (TCMIH) proposal/request for donation', is_interesting=False),
    DocCfg(id='029448', description=f"weird short essay titled 'President Obama and Self-Deception'"),

    # DOJ files
    DocCfg(
        id='EFTA00034357',
        author=BUREAU_OF_PRISONS,
        date='2019-08-10',
        description=f"internal message about discovery of Epstein's body",
        background_color='red'
    ),
    DocCfg(id='EFTA01103465', author=BEN_GOERTZEL, date='2013-12-02', description='funding proposal for AI labs in Africa etc.'),
    DocCfg(id='EFTA00165515', description="contractor describes Epstein's gun safes", show_full_panel=True),
    DocCfg(id='EFTA00266322', description=f"documents about pitches for non-profits in Australia, including to Effective Altruism"),
    DocCfg(id='EFTA00029538', description=f"{GHISLAINE_MAXWELL} provided school girl uniforms", show_full_panel=True),
    DocCfg(id='EFTA00005783', description='heavily redacted handwritten note, 30+ completely redacted pages', date='2019-08-29'),
    DocCfg(id='EFTA00005386', description='heavily redacted photo album, lot of photos of girls'),
    DocCfg(id='EFTA01063691', description='inventory of address books and Skype logs seized from Epstein computers'),
    DocCfg(id='EFTA00024275', description='large Wexner funded payments to OB-GYN'),
    DocCfg(id='EFTA00728783', description='list of names and phone numbers'),
    DocCfg(id='EFTA00298379', description='RSVP list for Yom Kippur dinner', date='2010-09-18'),
    # Urramoor
    DocCfg(
        id='EFTA01107738',
        description=f"creation of {CANTOR_URRAMOOR} with Mr. T's (Prince Andrew?) Urramoor and {LUTNICKS_CANTOR}",
        is_interesting=True,
    ),
    DocCfg(
        id='EFTA01141453',
        description=f"referral agreement between Mr. T's (Prince Andrew?) Urramoor and {LUTNICKS_CANTOR}",
        is_interesting=True,
    ),
    # Replacement text
    DocCfg(
        id='EFTA00009622',
        date='2006-07-19',
        description='handwritten notes from a victim interview transcribed by Claude AI',
        is_interesting=True,
        replace_text_with=EFTA00009622_TEXT,
    ),
    DocCfg(
        id='EFTA00004477',
        is_interesting=True,
        replace_text_with='Epstein 50th birthday photo book 12 "THAIS, MOSCOW GIRLS, AFRICA, HAWAII, [REDACTED] [REDACTED], Zorro, [REDACTED] [REDACTED] [REDACTED], CRACK WHOLE PROPOSAL, BALI/THAILAND/ASIA, RUSSIA, [REDACTED], [REDACTED], NUDES, YOGAL GIRLS',
    ),
    DocCfg(id='EFTA01628970', replace_text_with='redacted pictures of naked women'),
    DocCfg(id='EFTA00004070', replace_text_with="photos of Epstein with handwritten caption that didn't OCR well"),
    DocCfg(id='EFTA02731260', replace_text_with='notebook full of handwritten love letters with terrible OCR text'),
    DocCfg(id='EFTA00006100', replace_text_with='Palm Beach Police fax machine activity log 2005-12-28 to 2006-01-04'),
    DocCfg(
        id='EFTA00003149',
        date='2016-01-01',
        date_uncertain='complete guess',
        description=f"{LITTLE_SAINT_JAMES} staff list",
        show_full_panel=True,
    ),
    # Attachments
    DocCfg(id='EFTA00590685', attached_to_email_id='EFTA00830911'),
    # Dates
    DocCfg(id='EFTA02025218', date='2011-09-09'),
    # FedEx
    fedex_invoice('EFTA00217072', '2005-06-20'),
    fedex_invoice('EFTA00217080', '2005-06-27'),
    # Message pads
    important_messages_pad('EFTA01719859', '2005-10-03'),
    important_messages_pad('EFTA01682477', '2005-04-01'),
]



import re
from copy import copy
from dataclasses import dataclass, field
from os import devnull
from pathlib import Path
from typing import Literal, Mapping

from rich.console import Console, RenderableType
from rich.panel import Panel
from rich.style import Style
from rich.text import Text

from epstein_files.output.rich import CONSOLE_KWARGS, RICH_THEME
from epstein_files.util.constant.html import FONT_FAMILY, HTML_TERMINAL_THEME
from epstein_files.util.helpers.data_helpers import sort_dict_by_keys
from epstein_files.util.helpers.string_helper import quote
from epstein_files.util.logging import logger

CssProps = dict[str, str] | None
CssUnit = str | int
Side = Literal['top', 'left', 'right', 'bottom']
SideProp = Literal['margin', 'padding']

HORIZONTAL_SIDES: list[Side] = ['left', 'right']
VERTICAL_SIDES: list[Side] = ['top', 'bottom']
ALL_SIDES: list[Side] = VERTICAL_SIDES + HORIZONTAL_SIDES

CODE_TEMPLATE = '{code}'
SPLITTER = '-# JUNK #-'
SPLITTER_TEMPLATE = SPLITTER + """{stylesheet} {background} {foreground}"""  # these template vars allow export_html() to work

# CSS classes
BLACK_BACKGROUND = 'black_background'

# CSS dicts
CODE_TAG_CSS = {'font-family': 'inherit'}
FONT_CSS_PROPS = {'font-family': FONT_FAMILY}
HTML_CONSOLE_KWARGS = copy(CONSOLE_KWARGS)
HTML_CONSOLE_KWARGS.update({'file': open(devnull, "wt"), 'record': True})
PRE_TAG_CSS = {}

html_console = Console(**HTML_CONSOLE_KWARGS)


@dataclass
class HtmlStyle:
    _style: Style | str | None
    style: Style = field(init=False)

    def __post_init__(self):
        if isinstance(self._style, Style):
            self.style = self._style
        elif self._style:
            if self._style in RICH_THEME.styles:
                self.style = RICH_THEME.styles[self._style]
            else:
                self.style = Style.parse(self._style)
        else:
            self.style = Style.parse('')

    @property
    def bg_hex(self) -> str:
        if self.style.bgcolor:
            return self.style.bgcolor.get_truecolor(HTML_TERMINAL_THEME).hex
        else:
            return ''

    @property
    def hex(self) -> str:
        if self.style.color:
            return self.style.color.get_truecolor(HTML_TERMINAL_THEME).hex
        else:
            return ''

    @property
    def to_css(self) -> dict[str, str]:
        props = {}

        if self.bg_hex:
            props['background-color'] = self.bg_hex
        if self.hex:
            props['color'] = self.hex

        return props


def div_tag(contents: str, css_props: CssProps = None, **kwargs) -> str:
    return tag('div', contents, css_props, **kwargs)


def div_class(contents: str, class_name: str, css_props: CssProps = None, **kwargs) -> str:
    return div_tag(contents, css_props, class_name=class_name, **kwargs)


def div_with_legend(contents: str, legend: str, css_props: CssProps = None, **kwargs) -> str:
    """Use <fieldset> + <legend>: https://css-tricks.com/how-to-add-text-in-borders-using-basic-html-elements/"""
    css_props = css_props or {}

    if legend:
        legend_css_props = {'color': css_props.get('border-color', 'white')}
        contents = tag('legend', legend, legend_css_props) + contents

    return tag('fieldset', contents, css_props)


def from_em(units: str | None) -> int | None:
    if not units:
        return None

    return int(units.removesuffix('em'))


def horizontal_margin_props(units: CssUnit) -> dict[str, str]:
    return side_props('margin', HORIZONTAL_SIDES, units)


def horizontal_pad_props(units: CssUnit) -> dict[str, str]:
    return side_props('padding', HORIZONTAL_SIDES, units)


def list_tag(elements: list[str], list_tag: Literal['ol', 'ul'] = 'ul', **kwargs) -> str:
    if not elements:
        logger.warning(f"No elements to make <{list_tag}> for...")
        return ''

    txt_htmls = [tag('li', t) for t in elements]
    return tag(list_tag, '\n'.join(txt_htmls), **kwargs)


def margin_props(horizontal: CssUnit, vertical: CssUnit) -> dict[str, str]:
    return {**vertical_margin_props(horizontal), **vertical_margin_props(vertical)}


def padding_props(horizontal: CssUnit, vertical: CssUnit) -> dict[str, str]:
    return {**horizontal_pad_props(horizontal), **vertical_pad_props(vertical)}


def side_props(prop: SideProp, sides: list[Side], units: CssUnit) -> dict[str, str]:
    return {f"{prop}-{side}": to_px(units) for side in sides}


def strip_outer_tag(_html: str, tag: str) -> str:
    html = re.sub(fr"^\s*<{tag}.*?>", '', _html)
    html = re.sub(fr"</{tag}>\s*$", '', html)

    if html == _html:
        logger.warning(f"asked to strip <{tag}> from {html} but nothing stripped...")

    return html


def tag(tag: str, contents: str, css_props: CssProps = None, **kwargs) -> str:
    """Surround `contents` with <tag style="[CSS_STRING]">."""
    tag_kwargs = f' style="{to_inline_css(css_props)}"' if css_props else ''

    if kwargs:
        if 'class_name' in kwargs:
            kwargs['class'] = kwargs.pop('class_name')

        tag_kwargs += ' ' + ' '.join([f'{k}={quote(v)}' for k, v in kwargs.items()])

    return f'<{tag}{tag_kwargs}>{contents}</{tag}>'


def to_inline_css(d: Mapping[str, str | int]) -> str:
    css_props = [f'{k}: {v}' for k, v in sort_dict_by_keys(d).items() if v]
    return '; '.join(css_props)


def to_em(chars: int | float | None) -> str:
    return f"{chars}em" if chars else ''


def to_px(pixels: CssUnit) -> str:
    return f"{pixels}px" if isinstance(pixels, int) else pixels


def vertical_margin_props(units: CssUnit) -> dict[str, str]:
    return side_props('margin', VERTICAL_SIDES, units)


def vertical_pad_props(units: CssUnit) -> dict[str, str]:
    return side_props('padding', VERTICAL_SIDES, units)


PRE_CONSOLE_TEMPLATE_PREFIX = tag('pre', CODE_TEMPLATE, PRE_TAG_CSS)
PRE_CONSOLE_TEMPLATE = PRE_CONSOLE_TEMPLATE_PREFIX + SPLITTER_TEMPLATE



import re
from pathlib import Path
from subprocess import check_output

from epstein_files.util.constant.strings import (DOJ_FILE_STEM_REGEX, DOJ_FILE_NAME_REGEX, EFTA_PREFIX,
     HOUSE_OVERSIGHT_NOV_2025_FILE_NAME_REGEX, HOUSE_OVERSIGHT_NOV_2025_FILE_STEM_REGEX,
     HOUSE_OVERSIGHT_NOV_2025_ID_REGEX, HOUSE_OVERSIGHT_PREFIX, LOCAL_EXTRACT_REGEX)
from epstein_files.util.env import DOCS_DIR, DOJ_TXTS_20260130_DIR, DROPSITE_EMLS_DIR
from epstein_files.util.logging import logger

EXTRACTED_EMAILS_DIR = Path('emails_extracted_from_legal_filings')
DOJ_FILE_ID_REGEX = re.compile(fr".*{DOJ_FILE_NAME_REGEX.pattern}")
HOUSE_FILE_ID_REGEX = re.compile(fr".*{HOUSE_OVERSIGHT_NOV_2025_FILE_NAME_REGEX.pattern}")

DROPSITE_FILE_NAME_REGEX = re.compile(fr"{DROPSITE_EMLS_DIR}.* (\d\d\d\d-\d\d-\d\d \d+)\.eml")
FILENAME_LENGTH = len(HOUSE_OVERSIGHT_PREFIX) + 6
KB = 1024
MB = KB * KB

all_txt_paths = lambda: doj_txt_paths() + oversight_txt_paths() + dropsite_eml_paths()
doj_txt_paths = lambda: [f for f in DOJ_TXTS_20260130_DIR.glob('**/*.txt')] if DOJ_TXTS_20260130_DIR else []
dropsite_eml_paths = lambda: [f for f in DROPSITE_EMLS_DIR.glob('*.eml')]
oversight_txt_paths = lambda: [f for f in DOCS_DIR.iterdir() if f.is_file() and not f.name.startswith('.')]

file_size = lambda file_path: Path(file_path).stat().st_size


# Coerce methods handle both string and int arguments.
def coerce_file_name(filename_or_id: int | str | Path) -> str:
    return coerce_file_stem(filename_or_id) + '.txt'


def coerce_file_path(filename_or_id: int | str) -> Path:
    """Returns the `Path` for the file with `filename_or_id` ID."""
    filename = coerce_file_name(filename_or_id)

    if isinstance(filename_or_id, str) and is_doj_file(filename_or_id):
        for txt_file in DOJ_TXTS_20260130_DIR.glob('**/*.txt'):
            if txt_file.name == filename:
                return txt_file

        raise FileNotFoundError(f"'{filename_or_id}' looks like a DOJ file ID but no file named {filename} in '{DOJ_TXTS_20260130_DIR}'!")
    else:
        return DOCS_DIR.joinpath(filename)


def coerce_file_stem(filename_or_id: int | str | Path) -> str:
    """Generate a valid file stem no matter what form the argument comes in."""
    if isinstance(filename_or_id, str) and is_doj_file(filename_or_id):
        return Path(filename_or_id).stem
    elif isinstance(filename_or_id, Path):
        return filename_or_id.stem

    # print(f"\n\n, filename_or_id={filename_or_id} ({type(filename_or_id)})")
    # import pdb;pdb.set_trace()

    if isinstance(filename_or_id, str) and filename_or_id.startswith(HOUSE_OVERSIGHT_PREFIX):
        file_id = extract_file_id(filename_or_id)
        file_stem = file_stem_for_id(file_id)
    else:
        file_stem = file_stem_for_id(str(filename_or_id))

    if not HOUSE_OVERSIGHT_NOV_2025_FILE_STEM_REGEX.match(file_stem):
        raise RuntimeError(f"Invalid stem '{file_stem}' from '{filename_or_id}'")

    return file_stem


def coerce_url_slug(filename_or_id: int | str | Path) -> str:
    file_stem = coerce_file_stem(filename_or_id)

    if is_local_extract_file(file_stem):
        return LOCAL_EXTRACT_REGEX.sub('', file_stem)
    else:
        return file_stem


def extract_efta_id(file_id: str) -> int:
    return int(file_id.removeprefix(EFTA_PREFIX))


def extract_file_id(filename_or_id: int | str | Path) -> str:
    # DOJ 2026-01 files have different pattern
    if isinstance(filename_or_id, (str, Path)) and is_doj_file(filename_or_id):
        return Path(filename_or_id).stem
    elif isinstance(filename_or_id, (str, Path)) and (m := DROPSITE_FILE_NAME_REGEX.match(str(filename_or_id))):
        return f"DropSite {m.group(1)}"
    elif isinstance(filename_or_id, str):
        filename_or_id = filename_or_id.removesuffix(',')  # clean up commas from bad args.positional_args copypasta

    if isinstance(filename_or_id, int) or (isinstance(filename_or_id, str) and len(filename_or_id) <= 6):
        return format_house_oversight_id(filename_or_id)

    filename_str = str(filename_or_id)

    if (file_match := HOUSE_OVERSIGHT_NOV_2025_ID_REGEX.search(filename_str.upper())):
        return file_match.group(1)
    else:
        logger.error(f"filename_str='{filename_str}', HOUSE_FILE_ID_REGEX='{HOUSE_FILE_ID_REGEX.pattern}'")
        raise RuntimeError(f"Failed to extract file ID from '{filename_or_id}' (type: {type(filename_or_id).__name__}!")


def file_size_str(file_path: str | Path, digits: int | None = None):
    return file_size_to_str(file_size(file_path), digits)


def file_size_to_str(size: int, digits: int | None = None) -> str:
    _digits = 2

    if size > MB:
        size_num = float(size) / MB
        size_str = 'MB'
    elif size > KB:
        size_num = float(size) / KB
        size_str = 'kb'
        _digits = 1
    else:
        return f"{size} b"

    digits = _digits if digits is None else digits
    return f"{size_num:,.{digits}f} {size_str}"


def file_stem_for_id(id: int | str) -> str:
    if isinstance(id, int) or (isinstance(id, str) and len(id) <= 6):
        return f"{HOUSE_OVERSIGHT_PREFIX}{format_house_oversight_id(id)}"
    elif HOUSE_OVERSIGHT_NOV_2025_ID_REGEX.match(id):
        return f"{HOUSE_OVERSIGHT_PREFIX}{id}"
    else:
        raise ValueError(f"Unknown kind of file id {id}")


def format_house_oversight_id(id: int | str) -> str:
    """Make sure there's enough leading zeroes for 6 digit ID."""
    return f"{int(id):06d}"


def is_doj_file(file: str | Path) -> bool:
    """Check for EFTAXXXXXXX style files."""
    return bool(DOJ_FILE_STEM_REGEX.search(str(file)))


def is_house_oversight_file(file: str | Path) -> bool:
    return bool(HOUSE_OVERSIGHT_NOV_2025_FILE_STEM_REGEX.search(str(file)))


def is_local_extract_file(filename: str | Path) -> bool:
    """Return True if `filename` is of form 'HOUSE_OVERSIGHT_029835_1.txt'."""
    match = HOUSE_FILE_ID_REGEX.search(str(filename)) or DOJ_FILE_ID_REGEX.search(str(filename))
    return True if match and match.group(2) else False


def log_file_write(file_path: str | Path) -> None:
    logger.warning(f"Wrote {file_size_str(file_path)} to '{file_path}'")


def open_file_or_url(thing_to_open: str | Path) -> None:
    cmd = 'code' if str(thing_to_open).endswith('.txt') else 'open'
    check_output([cmd, str(thing_to_open)])
